package uoc.ei.exemples.modul9.dns;

import java.io.*; // necesari per a les operacions d'entrada i sortida
import java.util.StringTokenizer; // ajuda per interpretar les entrades

import uoc.ei.tads.Iterador;

/**
 * Classe principal encarregada d'executar un joc de proves provinent d'un
 * fitxer o de l'entrada estndard. Els resultats s'escriuen a un fitxer o
 * a la sortida estndard.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class Principal {

	BufferedReader in;
	PrintStream out;
	DNS dns;


   /**
    * Constructor que rep la llista d'arguments de la lnia de comanda.
    *
    * @param args  arguments de la lnia de comandes.<br>
    * Podem rebre diferents nombres de parmetres:<br>
    * -cap (utilitzarem l'entrada i sortida estndard)<br>
    * -un (utilitzem l'entrada estndard i la sortida cap a fitxer args[0])<br>
    * -dos (entrada del fitxer args[0] i sortida al fitxer args[1]<br>
    */
 	public Principal(String[] args) throws FileNotFoundException {
		switch (args.length) {
			case 0:
				in=new BufferedReader(new InputStreamReader(System.in));
				out=System.out;
				break;
			case 1:
				in=new BufferedReader(new FileReader(args[0]));
				out=System.out;
				break;
			case 2:
				in=new BufferedReader(new FileReader(args[0]));
				out=new PrintStream(new FileOutputStream(args[1]));
				break;
			default:
				throw new NullPointerException("ERROR en el nombre de parmetres:"+args.length);
		}
	}


   /**
    * Destructor que s'encarrega de tancar el fitxer d'entrada i sortida.
    */
	protected void finalize() throws IOException {
		in.close();
		out.close();
	}


   /**
    * Mtode que crea i inicialitza un nou objecte DNS
    */
	void creaDNS() {
		dns=new DNSimpl();
		out.println("DNS CREAT");
	}


   /**
    * Mtode que crea i inicialitza un nou objecte DNS
    *
    * @param domini cadena que cont el domini. ex: "www.uoc.edu".
    * @param ip  cadena que cont l'adrea ip del domini. ex: "129.34.2.3".
    */
	void creaDomini(String domini,String ip) {
		dns.altaDomini(new Domini(domini),new IP(ip));
		out.println("ALTA "+domini+"-->"+ip);
	}


   /**
    * Mtode que dna de baixa un domini.
    *
    * @param domini cadena que cont el domini. ex: "www.uoc.edu"
    */
	void baixaDomini(String domini) {
		dns.baixaDomini(new Domini(domini));
		out.println("BAIXA "+domini);
	}


   /**
    * Mtode que consulta l'adrea ip d'un domini.
    *
    * @param domini cadena que cont el domini. ex: "www.uoc.edu"
    */
	void consultaDomini(String domini) {
		IP ip=dns.consultar(new Domini(domini));
		if (ip!=null) out.println("CONSULTA "+domini+"-->"+ip);
		else out.println("CONSULTA "+domini+" inexistent");
	}


   /**
    * Mtode que consulta el domini ms visitat que ocupa la posici nom dins del ranking.
    *
    * @param num posici dins dels top 10. Ha d'anar de 0 a 9.
    */
	void mesVisitats(String num) {
		Domini d=dns.mesVisitats(Integer.parseInt(num));
		if (d!=null) out.println("MES VISITATS "+num+"-->"+d);
		else out.println("MES VISITATS "+num+" inexistent");
	}


	/**
    * Mtode que inicialitza un tld per a fer un recorregut per les seves entitats.
    *
    * @param tld cadena que cont el tld. ex: "edu"
    */
	void entitats(String tld) {
		Iterador<Entitat> iter=dns.entitats(new TLD(tld));
		out.print("ENTITATS "+tld+": ");
		if (iter!=null) {
			while (iter.hiHaSeguent()) {
				Entitat entitat=iter.seguent();
				System.out.print(entitat);
				if (iter.hiHaSeguent())
					System.out.print(", ");
			}
			out.println();
		}
		else
			out.println("tld no trobat!");
	}


	/**
    * Mtode que interpreta una lnia de test. Les comandes poden ser:
    * CREA
    * ALTA <domini> <ip>
    * BAIXA <domini>
    * CONSULTA <domini> --> <ip>
    * MESVISITATS <num> --> <domini>
    * ENTITATS <tld>
    * # 		//per afegir comentaris
    *
    * @param tld cadena que cont la comanda.
    */
	void interpretaLinia(String linia) {
	    StringTokenizer l=new StringTokenizer(linia," ");
	    if (l.hasMoreTokens()) {
	        String ins = l.nextToken();
            if (ins.equalsIgnoreCase("CREA")) {creaDNS();
            } else if (ins.equalsIgnoreCase("ALTA")) {creaDomini(l.nextToken(),l.nextToken());
            } else if (ins.equalsIgnoreCase("BAIXA")) {baixaDomini(l.nextToken());
            } else if (ins.equalsIgnoreCase("CONSULTA")) {consultaDomini(l.nextToken());
            } else if (ins.equalsIgnoreCase("MESVISITATS")) {mesVisitats(l.nextToken());
            } else if (ins.equalsIgnoreCase("ENTITATS")) {entitats(l.nextToken());
            } else if (ins.equals("#")) { System.exit(0);
            }
	    }
	}

   /**
    * Mtode que llegeix comandes de l'entrada definida (estndard o fitxer).
    *
    */
	public void executa() throws IOException{
		String linia;
		while ((linia = in.readLine())!=null && !linia.equals("FI")) {
            interpretaLinia(linia);
        }
	}

   /**
    * Programa principal.
    */
	public static void main(String[] args) throws IOException {
		Principal p=new Principal(args);
		p.executa();
		System.out.println("Final");
	}
}
