package uoc.ei.tads;
/**
 * Classe que implementa un arbre binari de cerca equilibrat AVL (Adelson-
 * Velskii & Landis), el qual es caracteritza per tenir l'arrel ms gran
 * que tots els elements del seu subarbre esquerre (si en t) i ms petita
 * que tots els elements del seu subarbre dret (si en t); i, a ms, els
 * seus subarbres esquerre i dret (si n'hi ha) sn tamb arbres de cerca.
 *
 * Es diu que l'arbre est equilibrat (height balanced) quan el valor
 * absolut de la diferncia d'altures dels seus subarbres s menor o igual
 * que u i els seus subarbres tamb estan equilibrats.
 *
 * S'espera que les classes dels elements implementin java.lang.Comparable
 * o b que es faciliti un java.util.Comparator com a parmetre del
 * constructor.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class ArbreAVL<E>  extends ArbreBinariCercaEncadenat<E> {
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


   /**
    * Constructor sense parmetres. S'espera que les classes dels elements
    * implementin la interfcie java.lang.Comparable.
    */
   public ArbreAVL() { super(); }

   /**
    * Constructor amb un parmetre i elements d'una classe comparable amb
    * el comparador donat.
    * @param comparador  comparador que permet deduir la prioritat
    * @exception ExcepcioParametreIncorrecte  si el comparador s null
    */
   public ArbreAVL(java.util.Comparator<E> comparador) {
      super(comparador);
   }

   /**
    * Mtode que restaura, si cal, l'equilibri de l'arbre de nodes
    * desprs de cada inserci o supressi. S'espera un NodeAVL no null.
    * @param node  arbre o subarbre de nodes que s'ha d'equilibrar
    */
   protected void equilibrar(Posicio<E> node) {
      if (node != null)  ((NodeAVL)node).equilibrar();
   }

   /**
    * Sobreescriu el mtode de la superclasse per a incorporar l'atribut
    * d'altura al node, que permet comprovar l'equilibri de l'arbre AVL.
    * @param pare  pare del nou node
    * @param elemComp  element comparable que es vol desar al node
    * @return  node arbre binari nou que emmagatzema l'element i incorpora
    *          l'altura per defecte u (fulla)
    */
   protected NodeArbre<E> crearNode(Posicio<E> pare, E elemComp) {
      return  new NodeAVL<E>(elemComp);
   }

   /**
    * Classe que estn NodeArbre per a incorporar un atribut d'altura que
    * permet comprovar i mantenir l'equilibri d'un arbre de nodes AVL.
    */
   protected static class NodeAVL<EN>  extends NodeArbre<EN> {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();


      /** Altura del node dins de l'arbre. Per defecte val u (fulla). */
      protected int altura = 1;

      /**
       * Constructor amb un parmetre. Assigna el valor rebut a l'element
       * del node pare i dna valor null a les posicions filles.
       * Delega en la superclasse. L'altura per defecte val u (fulla).
       * @param elem  valor de l'element que ha d'anar al node
       */
      public NodeAVL(EN elem) {
      	super();					// Sembla haver-hi bug en JDK 1.5
      	element=elem;				// que no deixa fer servir super(elem)
      }

      /**
       * Accessor d'escriptura de l'altura continguda al node.
       * @param altura  nou valor de l'altura
       */
      public void setAltura(int altura) { this.altura = altura; }

      /**
       * Accessor de lectura de l'altura continguda al node.
       * @return  altura del node
       */
      public int getAltura() { return  altura; }

      /**
       * Mtode que actualitza l'altura del node. Suma u (node pare) al
       * valor de l'altura del seu fill ms alt.
       */
      public void ajustarAltura() {
         int fe = (fillEsquerre == null) ? 0 :
                  ((NodeAVL<EN>)fillEsquerre).altura;
         int fd = (fillDret == null) ? 0 : ((NodeAVL<EN>)fillDret).altura;
         altura = 1 + ((fe > fd) ? fe : fd);
      }

      /**
       * Mtode que comprova l'equilibri de l'arbre de nodes.
       * @return  zero si els seus fills tenen la mateixa altura;
       *          negatiu si penja cap a la dreta;
       *          i positiu si el subarbre esquerre s el ms alt
       */
      public int balanceig() {
         int fe = (fillEsquerre == null) ? 0 :
                  ((NodeAVL<EN>)fillEsquerre).altura;
         int fd = (fillDret == null) ? 0 : ((NodeAVL<EN>)fillDret).altura;
         return  (fe - fd);
      }

      /**
       * Mtode que restaura, si cal, l'equilibri de l'arbre de nodes
       * desprs de cada inserci o supressi. El factor d'equilibri ha
       * d'estar entre -1 i 1. Es poden donar quatre tipus de desequilibri
       * (EE, ED, DD i DE) segons el balanceig (a l'esquerra o a la dreta)
       * de l'arbre i del subarbre corresponent.
       */
      public void equilibrar() {
         ajustarAltura();
         if (balanceig() > 1) {               // balancejat cap a l'esquerra
            if (((NodeAVL<EN>)fillEsquerre).balanceig() > -1)
				rotarEE();
            else
				rotarED();
         }
         else if (balanceig() < -1) {         // balancejat cap a la dreta
            if (((NodeAVL<EN>)fillDret).balanceig() < 1 )
				rotarDD();
            else
				rotarDE();
         }
      }


	  /**
       * Mtode que corregeix el desequilibri DD (dreta dreta).
       * El node s'ha inserit en el subarbre dret del subarbre dret, o b
       * s'ha suprimit un node del subarbre esquerre en les mateixes
       * circumstncies (en un arbre balancejat cap a la dreta).
       * s a dir, l'arbre ha crescut per (T2) o ha minvat per (T0).
       * L'arrel B del subarbre dret passa a ser la nova arrel i conserva
       * el seu fill dret, que amb aquest moviment queda a un nivell ms a
       * prop de l'arrel de l'arbre.
       * <PRE>
       *      A                                  B
       *    /   \                              /   \
       *  /       \                          /       \
       * T0        B       ---------->      A        T2
       *         /   \                    /   \
       *       /       \                /       \
       *      T1       T2              T0       T1
       * </PRE>
       * L'antiga arrel A esdev fill esquerre de B i conserva el seu
       * subarbre esquerre. Finalment, l'anterior subarbre esquerre de B
       * esdev subarbre dret de l'antiga arrel A per a conservar la
       * propietat d'ordenaci dels arbres de cerca.
       */
      protected void rotarDD() {
         NodeAVL<EN> node = (NodeAVL<EN>)fillEsquerre;
         fillEsquerre = fillDret;
         fillDret = fillEsquerre.fillDret;
         fillEsquerre.fillDret =
                      fillEsquerre.fillEsquerre;
         fillEsquerre.fillEsquerre = node;

         EN elemComp = element;
         element = fillEsquerre.element;
         fillEsquerre.element = elemComp;

         // ajusta altura d'A (antiga arrel) i de B (nova arrel subarbre)
         ((NodeAVL<EN>)fillEsquerre).ajustarAltura();
         ajustarAltura();
      }

      /**
       * Mtode protegit que corregeix el desequilibri DE (dreta
       * esquerra). El node s'ha inserit en el subarbre esquerre del
       * subarbre dret, o b s'ha suprimit un node del subarbre esquerra
       * en les mateixes circumstncies (en un arbre balancejat cap a la
       * dreta).
       * s a dir, l'arbre ha crescut per (C) o ha minvat per (T0).
       * <PRE>
       *      A       ----->        A
       *    /   \                 /   \
       *  /       \             /       \
       * T0        B           T0        C        ----->        C
       *         /   \                 /   \                  /   \
       *       /       \             /       \              /       \
       *      C        T3           T1        B            A         B
       *     / \                             / \          / \       / \
       *    /   \                           /   \        /   \     /   \
       *   T1   T2                         T2   T3      T0   T1   T2   T3
       * </PRE>
       * Per a restaurar l'equilibri calen dues rotacions: primer es
       * corregeix el desequilibri EE del subarbre dret i, a continuaci,
       * el desequilibri DD de l'arbre.
       * @see NodeAVL#rotarEE()
       * @see NodeAVL#rotarDD()
       */
      protected void rotarDE() {
         ((NodeAVL<EN>)fillDret).rotarEE();
         rotarDD();
      }

      /**
       * Mtode que corregeix el desequilibri EE (esquerra esquerra).
       * El procediment s simtric al cas del desequilibri DD.
       * <PRE>
       *            A                                  B
       *          /   \                              /   \
       *        /       \                          /       \
       *       B        T0       ---------->      T2        A
       *     /   \                                        /   \
       *   /       \                                    /       \
       *  T2       T1                                  T1       T0
       * </PRE>
       * @see  NodeAVL#rotarDD()
       */
      protected void rotarEE() {
         NodeAVL<EN> node = (NodeAVL<EN>)fillDret;
         fillDret = fillEsquerre;
         fillEsquerre = fillDret.fillEsquerre;
         fillDret.fillEsquerre = fillDret.fillDret;
         fillDret.fillDret = node;

         EN elemComp = element;
         element = fillDret.element;
         fillDret.element = elemComp;

         ((NodeAVL<EN>)fillDret).ajustarAltura();
         ajustarAltura();
      }

      /**
       * Mtode protegit que corregeix el desequilibri ED (esquerra
       * dreta).
       * El procediment s simtric al cas del desequilibri DE.
       * <PRE>
       *            A        ----->        A
       *          /   \                  /   \
       *        /       \              /       \
       *       B        T0            C        T0    ----->     C
       *     /   \                  /   \                     /   \
       *   /       \              /       \                 /       \
       *  T3        C            B        T1               B         A
       *           / \          / \                       / \       / \
       *          /   \        /   \                     /   \     /   \
       *         T2   T1      T3   T2                   T3   T2   T1   T0
       * </PRE>
       * @see  NodeAVL#rotarDE()
       */
      protected void rotarED() {
         ((NodeAVL<EN>)fillEsquerre).rotarDD();
         rotarEE();
      }
   }
}
