package uoc.ei.tads;

/**
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public abstract class ArbreAbstracte<E> implements Arbre<E> {


	/**
	 * Comprova si l'arbre o subarbre t algn fill.
	 * @param node  posici de referncia
	 * @return cert si la posici t algun fill i fals en cas contrari
	 */
	public boolean esFulla(Posicio<E> node) {
		return !fills(node).hiHaSeguent();
	}


	public boolean estaBuit() {
		return arrel()==null;
	}


	/**
	 * Mtode que retorna el nombre de fills d'un node.
	 * @param node  posici de referncia
	 * @exception ExcepcioPosicioInvalida  si la posici s null o no
	 * vlida
	 * @return  quantitat de fills, o zero si s una fulla
	 */
	public int nombreFills(Posicio<E> node) {
		int n=0;
		Recorregut<E> fills=fills(node);
		while (fills.hiHaSeguent()) {
			n++;
			fills.seguent();
		}
      return  n;
	}


	/**
	* Accessor de lectura dels elements que hi ha al contenidor.
	* Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
	* lnies de codi: <PRE>
	* for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
	*    System.out.println(it.seguent()); </PRE>
	* Enumerar s simplement enunciar l'una darrere l'altra (les coses
	* d'una srie, les parts d'un tot). Per si el contenidor t definit
	* algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
	* conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
	* etc.), sense alterar l'estat actual del contenidor.
	* @see  Iterador#hiHaSeguent()
	* @see  Iterador#seguent()
	* @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
	*            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
	*            ms
	* @return  enumeraci dels elements del contenidor preordre
	*/
	public Iterador<E> elements() {
	      return  new IteradorRecorregutImpl<E,E>(recorregutPreordre());
	   }

	/**
	* Mtode que soporta mltiples recorreguts, de les posicions del
	* contenidor, simultanis i independents entre ells.
	* @return  enumeraci de les posicions del contenidor per nivell
	*/
	public Recorregut<E> posicions() { return  new RecorregutPerNivells<E>(this); }


	public Recorregut<E> recorregutPreordre() {
		return new RecorregutPreordre<E>(this);
	}


	public Recorregut<E> recorregutPostordre() {
		return new RecorregutPostordre<E>(this);
	}


	public Recorregut<E> recorregutPerNivells() {
		return new RecorregutPerNivells<E>(this);
	}


	public int nombreElems(Posicio<E> posicio) {
		int n=1;
		Recorregut<E> r=fills(posicio);
		while (r.hiHaSeguent())
			n+=nombreElems(r.seguent());
		return n;
	}


	public String toString() {
		StringBuffer sb=new StringBuffer("{");
		if (arrel()!=null)
			toString(sb,arrel());
		sb.append("}");
		return sb.toString();
	}


	protected void toString(StringBuffer sb,Posicio<E> posicio) {
		sb.append(posicio.toString()+" ");
		if (!esFulla(posicio)) {
			sb.append("{");
			Recorregut<E> r=fills(posicio);
			while (r.hiHaSeguent()) {
				toString(sb,r.seguent());
				if (r.hiHaSeguent())
					sb.append(",");
			}
			sb.append("}");
		}
	}


	/**
	 * Classe que proporciona el comportament bsic per als tres recorreguts
	 * preordre, inordre i postordre. En aquesta classe es defineix el
	 * comportament com a tots tres recorreguts, de manera que posteriorment
	 * nicament caldr redefinir, per a cada recorregut concret, el mtode
	 * segent.
	 * @see  Recorregut#hiHaSeguent()
	 * @see  Recorregut#seguent()
	 */
	protected static abstract class RecorregutOrdreBasic<E> implements Recorregut<E> {
		/** L'arbre que s'est recorrent. */
		protected Arbre<E> arbre;


		/** Pila auxiliar. Aquesta pila amagatzema els elements que hem tractat per que
		 * encara no hem recorregut. El cim de la pila sempre ha de tenir el segent element
		 * a recrrer. Tots tres recorreguts pre/in/post-ordre es poden veure de la segent
		 * manera: donat un node, aquest t un conjunt de descendents que apareixeran abans que
		 * ell en el recorregut (sn ms prioritaris) i un conjunt de descendents que apareixeran
		 * desprs d'ell en el recorregut (sn menys prioritaris). */
		protected Pila<Posicio<E>> pila;

		/** Constructor.
		 * @param arbre L'arbre a recrrer.
		 */
		public RecorregutOrdreBasic(Arbre<E> arbre) {
			this.arbre=arbre;
			pila = new PilaVectorImpl<Posicio<E>>(arbre.nombreElems());
			if (! arbre.estaBuit()) {
				pila.empilar(arbre.arrel());
				empilaDescendentsAmbMesPrioritat(arbre.arrel());
		    }
		}

		/** Aquest mtode empila els descendents d'un node que han d'aparixer
		 * abans que ell en el recorregut de l'arbre. Aquesta caracterstica
		 * diferncia els tres recorreguts preordre, inordre i
		 * postordre.
		 * Per tant, el mtode s definit en aquesta classe com a abstracte
		 * i ha de ser redefinit en les classes derivades de la manera pertinent.
		 * La resta de mtodes que implementen el recorregut es poden definir en
		 * aquesta mateixa classe en base a aquest mtode abstracte.
		 */
		protected abstract void empilaDescendentsAmbMesPrioritat(Posicio<E> pare);


		/** Aquest mtode empila els descendents d'un node que han d'aparixer
		 * desprs que ell en el recorregut de l'arbre. Aquesta caracterstica
		 * s l'altre caracterstica que diferncia els tres recorreguts preordre,
		 * inordre i postordre.
		 * Per tant, el mtode s definit en aquesta classe com a abstracte
		 * i ha de ser redefinit en les classes derivades de la manera pertinent.
		 * La resta de mtodes que implementen el recorregut es poden definir en
		 * aquesta mateixa classe en base a aquest mtode abstracte.
		 */
		protected abstract void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare);


		/** Aquest mtode empila els fills d'un node en l'ordre en el que estan definits.
		 * @param pare
		 */

		protected void empilaFills(Posicio<E> pare) {
			Recorregut<E> fills=arbre.fills(pare);
			// primer creem una pila amb els fills (per a poder-los recrrer en ordre invers)
			// la pila ha de ser encadenada, doncs no tenim fita. Com no tenim pila encadenada
			// fem servir directament una llista encadenada
			LlistaEncadenada<Posicio<E>> pilaFills=new LlistaEncadenada<Posicio<E>>();
			while (fills.hiHaSeguent()) {
				Posicio<E> fill=fills.seguent();
				pilaFills.afegirAlPrincipi(fill);
			}
			// un cop tenim els fills en ordre invers, els inserim a la pila del recorregut
			Iterador<Posicio<E>> fills2=pilaFills.elements();
			while (fills2.hiHaSeguent()) {
				Posicio<E> fill=fills2.seguent();
				pila.empilar(fill);
			}
		}


		/**
		 * Comprova si hi ha una primera o segent posici. s sensible a
		 * eventuals alteracions de l'estructura de posicions. Retorna fals
		 * si el contenidor est buit o ja s'ha visitat l'ltima posici.
		 * @return  cert o fals, segons si es pot avanar o no es pot
		 */
		public boolean hiHaSeguent() { return  ! pila.estaBuit(); }


		/**
		 * Primer avana, si es pot, i desprs retorna la posici.
		 * Si no hi ha segent posici llana una excepci.
		 * @return  segent posici
		 * @pre hiHaSeguent(), ExcepcioPosicioInvalida
		 */
		public Posicio<E> seguent() {
			Posicio<E> node=(Posicio<E>)pila.desempilar();
			empilaDescendentsAmbMenysPrioritat(node);
			return node;
		}
	}

	/**
	 * Classe que proporciona un recorregut de les posicions. Basada en el
	 * pattern Iterator, soporta mltiples recorreguts simultanis i
	 * independents del contenidor. s sensible a eventuals alteracions de
	 * l'estructura de posicions.
	 * @see  Recorregut#hiHaSeguent()
	 * @see  Recorregut#seguent()
	 */
	protected static class RecorregutPreordre<E> extends RecorregutOrdreBasic<E> {
		/** Atribut que en Java 1.5 determina la compatibilitat entre objectes
		 * serialitzables de la mateixa classe. L'deitificador es calcula
		 * mitjanant un mtode de la classe Utilitats. */
		private static final long serialVersionUID = Utilitats.getSerialVersionUID();


		/** Constructor.
		 * @param arbre L'arbre a recrrer.
		 */
		public RecorregutPreordre(Arbre<E> arbre) {
			super(arbre);
		}


		/** En el cas del recorregut preordre, donat un node, aquest sempre
		 * apareixer abans que els els seus descendents en el recorregut,
		 * per tant, un node no t descendents amb ms prioritat que ell.
		 */
		protected void empilaDescendentsAmbMesPrioritat(Posicio<E> pare) {}


		/** En el cas del recorregut preordre, els descendents amb menys
		 * prioritat que un node sn els seus dos fills (i a la vegada els
		 * descendents d'aquests). En tenim prou amb empilar els dos fills,
		 * doncs posteriorment quan arribem en el recorregut a aquests fills
		 * ja els tractarem i ja tractarem els seus fills adequadament.
		 */
		protected void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare) {
			empilaFills(pare);
		}
	}

	/**
	 * Classe que proporciona un recorregut de les posicions. Basada en el
	 * pattern Iterator, soporta mltiples recorreguts simultanis i
	 * independents del contenidor. s sensible a eventuals alteracions de
	 * l'estructura de posicions.
	 * @see  Recorregut#hiHaSeguent()
	 * @see  Recorregut#seguent()
	 */
	protected static class RecorregutPostordre<E> extends RecorregutOrdreBasic<E> {
		/** Atribut que en Java 1.5 determina la compatibilitat entre objectes
		 * serialitzables de la mateixa classe. L'deitificador es calcula
		 * mitjanant un mtode de la classe Utilitats. */
		private static final long serialVersionUID = Utilitats.getSerialVersionUID();


	      /** Darrer node del qual s'han empilat a la pila feta servir internament
	       * pel recorregut tant els fills amb ms prioritat com els fills amb menys
	       * prioritat */
	      private Pila<Posicio<E>> nodesTotalmentDesplegats;


		/** Constructor.
		 * @param arbre L'arbre a recrrer.
		 */
		public RecorregutPostordre(Arbre<E> arbre) {
			super(arbre);
		}


		/** En el cas del recorregut postordre, donat un node, aquest sempre
		 * apareixer desprs que els els seus descendents en el recorregut,
		 * per tant, tots els fills d'un node tenen ms prioritat que ell.
		 */
		protected void empilaDescendentsAmbMesPrioritat(Posicio<E> pare) {
			empilaFills(pare);
			while (!jaDesplegatTotalment(pila.cim()))
				empilaFills(pila.cim());
		}


		/** En el cas del recorregut postordre un node no t descendents amb
		 * menys prioritat.
		 */
		protected void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare) {}


		protected void empilaFills(Posicio<E> pare) {
			if (nodesTotalmentDesplegats==null)
				nodesTotalmentDesplegats=new PilaVectorImpl<Posicio<E>>(arbre.nombreElems());
			super.empilaFills(pare);
			nodesTotalmentDesplegats.empilar(pare);
		}


		public Posicio<E> seguent() {
	         Posicio<E> node=pila.desempilar();
	       	 nodesTotalmentDesplegats.desempilar();
	         if (!pila.estaBuit() && !jaDesplegatTotalment(pila.cim())) {
	        	 empilaDescendentsAmbMesPrioritat(pila.cim());
	         }
	         return node;
		}


		protected boolean jaDesplegatTotalment(Posicio<E> node) {
			return !nodesTotalmentDesplegats.estaBuit() && nodesTotalmentDesplegats.cim()==node;
		}

	}


	/**
	    * Classe que proporciona un recorregut de les posicions. Basada en el
	    * pattern Iterator, soporta mltiples recorreguts simultanis i
	    * independents del contenidor. s sensible a eventuals alteracions de
	    * l'estructura de posicions.
	    * @see  Recorregut#hiHaSeguent()
	    * @see  Recorregut#seguent()
	    */
	   protected static class RecorregutPerNivells<E>  implements Recorregut<E>
	   {
	      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
	       * serialitzables de la mateixa classe. L'deitificador es calcula
	       * mitjanant un mtode de la classe Utilitats. */
	      private static final long serialVersionUID = Utilitats.getSerialVersionUID();


	      /** Arbre que estem recorrent. */
	      Arbre<E> arbre;


	      /** Cua auxiliar. */
	      protected Cua<Posicio<E>> cua;


	      /** Constructor sense parmetres. */
	      public RecorregutPerNivells(Arbre<E> arbre)
	      {
	         this.arbre=arbre;
	         cua = new CuaVectorImpl<Posicio<E>>(arbre.nombreElems());
	         if (! arbre.estaBuit())  cua.encuar(arbre.arrel());
	      }

	      /**
	       * Comprova si hi ha una primera o segent posici. s sensible a
	       * eventuals alteracions de l'estructura de posicions. Retorna fals
	       * si est buit o ja s'ha visitat l'ltima posici.
	       * @return  cert o fals, segons si es pot avanar o no es pot
	       */
	      public boolean hiHaSeguent() { return  ! cua.estaBuit(); }

	      /**
	       * Primer avana, si es pot, i desprs retorna la posici.
	       * Si no hi ha segent posici llana una excepci.
	       * @return  segent posici
	       */
	      public Posicio<E> seguent()  throws  ExcepcioPosicioInvalida
	      {
	          if ( ! hiHaSeguent() )  throw new ExcepcioPosicioInvalida(
	                                           "No hi ha seguent");

	         Posicio<E> pare = cua.desencuar();      // desencua el pare
	         for (Recorregut<E> r = arbre.fills(pare); r.hiHaSeguent(); )
	         {
	            cua.encuar(r.seguent());                              // encua
	         }
	         return  pare;                                  // retorna el pare
	      }
	   }

}
