package uoc.ei.tads;
/**
 * Classe abstracta que defineix les operacions de qualsevol arbre binari,
 * el qual es caracteritza per organitzar els seus elements (nodes)
 * formant una jerarquia: tot node (tret de l'arrel que es el cap de la
 * jerarquia) s descendent d'un node nic, i pot ser ascendent d'un mxim
 * de dos nodes (quan no t descendents s'anomena fulla).
 *
 * Implementa els recorreguts habituals i altres operacions auxiliars que
 * es deriven de les operacions bsiques ja definides en aquest nivell de
 * la jerarquia.
 *
 * Implementa tamb la interfcie Serializable per a poder convertir a
 * cadenes o fluxes de bytes (streams) els objectes del contenidor i
 * gravar-los o transmetre'ls.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public abstract class ArbreBinari<E>  extends ArbreAbstracte<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /**
    * Afegeix un element com a fill esquerre de la posici rebuda, si es
    * pot.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @return  nou fill esquerre
    */
   public abstract Posicio<E> afegirFillEsquerre(Posicio<E> pare, E elem);

   /**
    * Afegeix un element com a fill dret de la posici rebuda, si es pot.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @return  nou fill dret
    */
   public abstract Posicio<E> afegirFillDret(Posicio<E> pare, E elem);

   /**
    * Accessor de lectura del fill esquerre d'una posici de l'arbre.
    * @param pos  posici de referncia
    * @return  fill esquerre de la posici rebuda; o null si no en t
    */
   public abstract Posicio<E> fillEsquerre(Posicio<E> pos);

   /**
    * Accessor de lectura del fill dret d'una posici de l'arbre.
    * @param pos  posici de referncia
    * @return  fill dret de la posici rebuda; o null si no en t
    */
   public abstract Posicio<E> fillDret(Posicio<E> pos);
   

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions filles
    * de la posici de referncia, simultanis i independents entre ells.
    * @param pare  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  enumeraci de les posicions filles
    */
   public Recorregut<E> fills(Posicio<E> pare) {
      return  new RecorregutFills<E>(this,pare);
   }

   /**
    * Comprova si l'arbre o subarbre t algn fill.
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  fals o cert, segons si t algun fill o no en t cap
    */
   public boolean esFulla(Posicio<E> node) {
      return  (fillEsquerre(node) == null) && (fillDret(node) == null);
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor inordre
    */
   public Iterador<E> elements() {
      return  new IteradorRecorregutImpl<E,E>(recorregutInordre());
   }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @see  RecorregutPerNivell
    * @return  enumeraci de les posicions del contenidor per nivell
    */
   public Recorregut<E> posicions() { return  new RecorregutPerNivell<E>(this); }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @see  RecorregutPreordre
    * @return  enumeraci de les posicions del contenidor preordre
    */
   public Recorregut<E> recorregutPreordre() {
      return  new RecorregutPreordre<E>(this);
   }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @see  RecorregutInordre
    * @return  enumeraci de les posicions del contenidor inordre
    */
   public Recorregut<E> recorregutInordre() {
      return  new RecorregutInordre<E>(this);
   }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @see  RecorregutPostordre
    * @return  enumeraci de les posicions del contenidor postordre
    */
   public Recorregut<E> recorregutPostordre() {
      return  new RecorregutPostordre<E>(this);
   }

   /**
    * Classe que proporciona un recorregut de les posicions filles. Basada
    * en el pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutFills<E>  implements Recorregut<E> {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
      
      /** Arbre que s'est recorrent. */
      ArbreBinari<E> arbre;
 	

      /** Cua per a emmagatzemar els fills. */
      private Cua<Posicio<E>> fills;

      /**
       * Constructor amb un parmetre.
       * @param pare  posici no null de referncia
       */
      public RecorregutFills(ArbreBinari<E> arbre,Posicio<E> pare) {
      	 this.arbre=arbre;
         fills = new CuaVectorImpl<Posicio<E>>(2);
         if (arbre.fillEsquerre(pare) != null)
             fills.encuar(arbre.fillEsquerre(pare));
         if (arbre.fillDret(pare) != null)
             fills.encuar(arbre.fillDret(pare));
      }

      /**
       * Comprova si hi ha una primera o segent posici. s sensible a
       * eventuals alteracions de l'estructura de posicions. Retorna fals
       * si s fulla o ja s'ha visitat l'ltima posici.
       * @return  cert o fals, segons si es pot avanar o no es pot
       */
      public boolean hiHaSeguent() { return  ! fills.estaBuit(); }

      /**
       * Primer avana, si es pot, i desprs retorna la posici.
       * Si no hi ha segent posici llana una excepci.
       * @return  segent posici
       * @pre hiHaSeguent(), ExcepcioPosicioInvalida
       */
      public Posicio<E> seguent()  throws  ExcepcioPosicioInvalida
      {
         return  fills.desencuar();
      }
   }

   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutPerNivell<E> implements Recorregut<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
      /** Arbre que s'est recorrent. */
      ArbreBinari<E> arbre;
 	

      /** Cua auxiliar. */
      protected Cua<Posicio<E>> cua;

      /** Constructor sense parmetres. */
      public RecorregutPerNivell(ArbreBinari<E> arbre)
      {
      	 this.arbre=arbre;
         cua = new CuaVectorImpl<Posicio<E>>(arbre.nombreElems());
         if (! arbre.estaBuit())  cua.encuar(arbre.arrel());
      }

      /**
       * Comprova si hi ha una primera o segent posici. s sensible a
       * eventuals alteracions de l'estructura de posicions. Retorna fals
       * si el contenidor est buit o ja s'ha visitat l'ltima posici.
       * @return  cert o fals, segons si es pot avanar o no es pot
       */
      public boolean hiHaSeguent() { return  ! cua.estaBuit(); }

      /**
       * Primer avana, si es pot, i desprs retorna la posici.
       * Si no hi ha segent posici llana una excepci.
       * @return  segent posici
       * @pre hiHaSeguent(), ExcepcioPosicioInvalida
       */
      public Posicio<E> seguent()  throws  ExcepcioPosicioInvalida
      {
         Posicio<E> pare = (Posicio<E>)cua.desencuar();
         if (arbre.fillEsquerre(pare) != null)
            cua.encuar(arbre.fillEsquerre(pare));
         if (arbre.fillDret(pare) != null)
            cua.encuar(arbre.fillDret(pare));
         return pare;
      }
   }

   
   /**
    * Classe que proporciona el comportament bsic per als tres recorreguts
    * preordre, inordre i postordre. En aquesta classe es defineix el
    * comportament com a tots tres recorreguts, de manera que posteriorment
    * nicament caldr redefinir, per a cada recorregut concret, el mtode
    * segent.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static abstract class RecorregutOrdreBasic<E> implements Recorregut<E>
   {
   	/** L'arbre que s'est recorrent. */
   	protected ArbreBinari<E> arbre;
   	
   	
    /** Pila auxiliar. Aquesta pila amagatzema els elements que hem tractat per que
     * encara no hem recorregut. El cim de la pila sempre ha de tenir el segent element
     * a recrrer. Tots tres recorreguts pre/in/post-ordre es poden veure de la segent
     * manera: donat un node, aquest t un conjunt de descendents que apareixeran abans que
     * ell en el recorregut (sn ms prioritaris) i un conjunt de descendents que apareixeran
     * desprs d'ell en el recorregut (sn menys prioritaris). */
    protected Pila<Posicio<E>> pila;

    /** Aquest mtode empila els descendents d'un node que han d'aparixer
     * abans que ell en el recorregut de l'arbre. Aquesta caracterstica
     * diferncia els tres recorreguts preordre, inordre i
     * postordre.
     * Per tant, el mtode s definit en aquesta classe com a abstracte
     * i ha de ser redefinit en les classes derivades de la manera pertinent.
     * La resta de mtodes que implementen el recorregut es poden definir en
     * aquesta mateixa classe en base a aquest mtode abstracte.
     */ 
    protected abstract void empilaDescendentsAmbMesPrioritat(Posicio<E> pare);

    
    /** Aquest mtode empila els descendents d'un node que han d'aparixer
     * desprs que ell en el recorregut de l'arbre. Aquesta caracterstica
     * s l'altre caracterstica que diferncia els tres recorreguts preordre,
     * inordre i postordre.
     * Per tant, el mtode s definit en aquesta classe com a abstracte
     * i ha de ser redefinit en les classes derivades de la manera pertinent.
     * La resta de mtodes que implementen el recorregut es poden definir en
     * aquesta mateixa classe en base a aquest mtode abstracte.
     */ 
    protected abstract void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare);


    /** Constructor.
     * @param arbre L'arbre a recrrer.
     */
    public RecorregutOrdreBasic(ArbreBinari<E> arbre)
    {
		this.arbre=arbre;
		pila = new PilaVectorImpl<Posicio<E>>(arbre.nombreElems());
		if (! arbre.estaBuit()) {
	       	pila.empilar(arbre.arrel());
	       	empilaDescendentsAmbMesPrioritat(arbre.arrel());
       }
    }


    /**
     * Comprova si hi ha una primera o segent posici. s sensible a
     * eventuals alteracions de l'estructura de posicions. Retorna fals
     * si el contenidor est buit o ja s'ha visitat l'ltima posici.
     * @return  cert o fals, segons si es pot avanar o no es pot
     */
    public boolean hiHaSeguent() { return  ! pila.estaBuit(); }
    

    /**
     * Primer avana, si es pot, i desprs retorna la posici.
     * Si no hi ha segent posici llana una excepci.
     * @return  segent posici
     * @pre hiHaSeguent(), ExcepcioPosicioInvalida
     */
    public Posicio<E> seguent()
    {
       if (pila.estaBuit())
    	   throw new ExcepcioPosicioInvalida();
       Posicio<E> node=(Posicio<E>)pila.desempilar();
       empilaDescendentsAmbMenysPrioritat(node);
       return node;
    }

   }

   
   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutPreordre<E> extends RecorregutOrdreBasic<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
      /** Constructor.
       * @param arbre L'arbre a recrrer.
       */
      public RecorregutPreordre(ArbreBinari<E> arbre)
      {
      	super(arbre);
      }

      	
      /** En el cas del recorregut preordre, donat un node, aquest sempre
	    * apareixer abans que els els seus descendents en el recorregut,
	    * per tant, un node no t descendents amb ms prioritat que ell.
	    */ 
	  protected void empilaDescendentsAmbMesPrioritat(Posicio<E> pare) {}
	

	   /** En el cas del recorregut preordre, els descendents amb menys
	    * prioritat que un node sn els seus dos fills (i a la vegada els
	    * descendents d'aquests). En tenim prou amb empilar els dos fills,
	    * doncs posteriorment quan arribem en el recorregut a aquests fills
	    * ja els tractarem i ja tractarem els seus fills adequadament.
	    */ 
	  protected void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare)
	   {
	   	Posicio<E> fill=arbre.fillDret(pare);
	   	if (fill!=null) pila.empilar(fill);
	   	fill=arbre.fillEsquerre(pare);
	   	if (fill!=null) pila.empilar(fill);
	   }
   }

   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutInordre<E> extends RecorregutOrdreBasic<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
      /** Constructor.
       * @param arbre L'arbre a recrrer.
       */
      public RecorregutInordre(ArbreBinari<E> arbre)
      {
      	super(arbre);
      }

      	
      /** Els descendents amb ms prioritat per al cas del recorregut
       * inordre sn el fill esquerre i els seus descendents. Per tant,
       * empilem el fill esquerre. Posteriorment, si el fill esquerre
       * t algun descendent amb ms prioritat, l'haurem d'empilar tamb.
       * Aix resulta en, en aquest cas, empilar els descendents esquerre;
       * es a dir, el fill esquerre, el fill esquerre del fill esquerre, etc.
       */ 
      protected void empilaDescendentsAmbMesPrioritat(Posicio<E> pare)
      {
      	Posicio<E> fill=arbre.fillEsquerre(pare);
  		while (fill!=null) {
  			pila.empilar(fill);
  			pare=fill;
  			fill=arbre.fillEsquerre(pare);
  		}
  	}

      /** Els descendents amb menys prioritat per al cas del recorregut
       * inordre sn el fill dret i els seus descendents. Per tant,
       * empilem el fill dret. Acte seguit passem a empilar els descendents
       * del fill dret que tenen ms prioritat que aquest.
       */ 
      protected void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare)
      {
      	Posicio<E> fill=arbre.fillDret(pare);
      	if (fill!=null) {
      		pila.empilar(fill);
      		empilaDescendentsAmbMesPrioritat(fill);
      	}
  	}

   }


   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutPostordre<E> extends RecorregutInordre<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
      
      
      /** Darrer node del qual s'han empilat a la pila feta servir internament
       * pel recorregut tant els fills amb ms prioritat com els fills amb menys
       * prioritat */
      private Pila<Posicio<E>> nodesTotalmentDesplegats; 
 	
 	
      /** Constructor.
       * @param arbre L'arbre a recrrer.
       */
      public RecorregutPostordre(ArbreBinari<E> arbre)
      {
       	super(arbre);
      }

      	
      /** Redefinici d'aquest mtode per a tenir en compte els descendents
       * amb menys prioritat del darrer node de la branca esquerra.
       */ 
      protected void empilaDescendentsAmbMesPrioritat(Posicio<E> pare)
      {
    	  super.empilaDescendentsAmbMesPrioritat(pare);
    	  empilaDescendentsAmbMenysPrioritat(pila.cim());
      }


      /** Els descendents amb menys prioritat per al cas del recorregut
       * inordre sn el fill dret i els seus descendents. Per tant,
       * empilem el fill dret. Acte seguit passem a empilar els descendents
       * del fill dret que tenen ms prioritat que aquest.
       */ 
      protected void empilaDescendentsAmbMenysPrioritat(Posicio<E> pare)
      {
    	  if (nodesTotalmentDesplegats==null)
    	       	nodesTotalmentDesplegats=new PilaVectorImpl<Posicio<E>>(arbre.nombreElems());
    	  nodesTotalmentDesplegats.empilar(pare);
    	  super.empilaDescendentsAmbMenysPrioritat(pare);
      }


      /** Aquesta classe reaprofita les prioritats establertes pel recorregut
       * inordre i modifica el mtode seguent. La modificaci consisteix en 
       * empilar els elements amb menys prioritat del node que queda al cim
       * de la pila desprs de desempilar l'actual.
       * s a dir, el que estem fent s canviar el significat dels mtodes
       * empilaDescendentsAmbMes/MenysPrioritat per a reaprofitar el mxim
       * codi possible. En aquesta classe tots els descendents d'un node
       * tenen ms prioritat que aquest. Llavors dividim els seus descendents
       * en dos grups: els que tenen ms prioritat (fill esquerre i descendents)
       * i els que en tenen menys (fill dret i descendents).
       * Si vulguessim respectar l'estructura definida a la classe base acabarem
       * tenint una pila que inicialment tindria tots els nodes de l'arbre.
       * @return  segent posici
       * @pre hiHaSeguent(), ExcepcioPosicioInvalida
       */
      public Posicio<E> seguent()
      {
         Posicio<E> node=pila.desempilar();
       	 nodesTotalmentDesplegats.desempilar();
         if (!pila.estaBuit() && !jaDesplegatTotalment(pila.cim())) {
        	 empilaDescendentsAmbMenysPrioritat(pila.cim());
         }
         return node;
      }
      
      
      protected boolean jaDesplegatTotalment(Posicio<E> node) {
    	  return !nodesTotalmentDesplegats.estaBuit() && nodesTotalmentDesplegats.cim()==node;
      }

   
   }

   
	protected void toString(StringBuffer sb,Posicio<E> posicio) {
		if (posicio!=null) {
			sb.append(posicio.toString()+" ");
			sb.append("{");
			toString(sb,fillEsquerre(posicio));
			sb.append(",");
			toString(sb,fillDret(posicio));
			sb.append("}");
		}
	}

}
