package uoc.ei.tads;
/**
 * Classe que implementa un arbre binari de cerca, el qual es caracteritza
 * per tenir l'arrel ms gran que tots els elements del seu subarbre
 * esquerre (si en t) i ms petita que tots els elements del seu subarbre
 * dret (si en t); i, a ms, els seus subarbres esquerre i dret (si n'hi
 * ha) sn tamb arbres de cerca.
 *
 * La classe implementa les operacions no posicionals d'afegir, esborrar i
 * consultar un element comparable, i noms t accessibilitat dins del
 * package perqu hereta unes operacions posicionals que podrien alterar
 * la consistncia de l'arbre de cerca. 
 *
 * S'espera que la classe dels elements implementin la interfcie
 * java.lang.Comparable o b que es faciliti un java.util.Comparator com a
 * parmetre del constructor.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
class ArbreBinariCercaEncadenat<E>  extends ArbreBinariEncadenatImpl<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /**
    * Comparador concret que permet deduir la prioritat entre els
    * elements. Pot tenir valor null i aleshores s'utilitza la interfcie
    * java.lang.Comparable
    */
   protected java.util.Comparator<E> comparador;

   /**
    * Constructor sense parmetres. S'espera que la classe dels elements
    * implementi la interfcie java.lang.Comparable.
    */
   public ArbreBinariCercaEncadenat() { super(); }

   /**
    * Constructor amb un parmetre i elements d'una classe comparable amb
    * el comparador donat.
    * @exception ExcepcioParametreIncorrecte  si el comparador s null
    * @param comparador  comparador que permet deduir la prioritat
    * @pre comparador!=null, ExcepcioParametreIncorrecte("comparador null")
    */
   public ArbreBinariCercaEncadenat(java.util.Comparator<E> comparador) {
      this.comparador = comparador;
   }

   /**
    * Afegeix un element a la posici que li correspon, si es pot. Si
    * l'element ja hi era, d'acord amb el comparador, el sobreescriu.
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @param elemComp  element comparable que es vol afegir a l'arbre
    */
   public void afegir(E elemComp) {
      afegirEnSubarbre(estaBuit() ? null : arrel(), elemComp);
   }

   /**
    * Esborra l'element, si el troba d'acord amb el comparador.
    * @exception ExcepcioContenidorBuit  si l'arbre est buit
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @param elemComp  element comparable que es vol suprimir de l'arbre
    * @return  l'element esborrat; o null, si no hi era
    */
   public E esborrar(E elemComp) {
      return  esborrarEnSubarbre(null, arrel(), elemComp);
   }

   /**
    * Accessor de lectura d'un element de l'arbre. Si no el troba, d'acord
    * amb el comparador, retorna null.
    * @exception ExcepcioContenidorBuit  si l'arbre est buit
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @param elemComp  element comparable que es vol consultar
    * @return  l'element interessat; o null, si no hi era
    */
   public E consultar(E elemComp) {
      return  consultarEnSubarbre(arrel(), elemComp);
   }

   /**
    * Mtode recursiu que garanteix la inserci ordenada dels elements.
    * No permet elements equivalents: si l'element existia, d'acord amb
    * l'operaci de comparaci, el sobreescriu; altrament crea un nou node
    * per a emmagatzemar l'element i l'encadena al lloc de l'arbre que li
    * correspon.
    * @param pare  posici de referncia (inicialment l'arrel)
    * @param elem  element comparable que es vol afegir a l'arbre
    * @return  nou fill; o l'arrel, si el pare era null
    */
   private Posicio<E> afegirEnSubarbre(Posicio<E> pare,E elem) {
      if (pare == null)
         return  super.afegir(pare, elem);
      Posicio<E> node = null;
      int comp = comparar(elem, pare.getElem());
      // depenent del resultat de comparar actuem. Hi ha tres casos diferents.
      // comp==0 (l'element s igual al pare), sobreescriu el pare
      if (comp == 0) {                                                  
         super.reemplacar(pare, elem);
         node=pare;
      }
      // comp<0: descarta l'arbre dret i continua l'afegir per l'esquerra
      else if (comp < 0) {
         if (fillEsquerre(pare) == null)
            node = afegirFillEsquerre(pare, elem);
         else
         	node=afegirEnSubarbre(fillEsquerre(pare), elem);
      }
      // comp>0: descarta esquerra i continua per dret
      else {
         if (fillDret(pare) == null)
            node = afegirFillDret(pare, elem);
         else
         	node=afegirEnSubarbre(fillDret(pare), elem);
      }
      equilibrar(pare);
      return node;
   }

   
   /**
    * Mtode recursiu que esborra l'element, si el troba d'acord amb el
    * comparador.
    * @param avi  posici de referncia (inicialment null)
    * @param pare  posici de referncia (inicialment l'arrel)
    * @param elem  element comparable que es vol suprimir de l'arbre
    * @return  element esborrat; o null, si no hi era
    */
   private E esborrarEnSubarbre(Posicio<E> avi, Posicio<E> pare, E elem) {
   	  E elemEsborrat=null;
      if (pare == null)
		  return  null;
      int comp = comparar(elem, pare.getElem());
      if (comp < 0)                                    // cap a l'esquerra
         elemEsborrat=esborrarEnSubarbre(pare, fillEsquerre(pare), elem);
      else if (comp > 0)                                 // cap a la dreta
         elemEsborrat=esborrarEnSubarbre(pare, fillDret(pare), elem);
      else {                                                  // 0: trobat
      	 elemEsborrat=pare.getElem();
         if (esFulla(pare))                              // si no t fills
            esborrar(avi, pare);                   // super.esborrar
         else if ( (fillEsquerre(pare) != null) &&
                   (fillDret(pare) == null) )        // noms t fill esq.
            reemplacarSubarbre(avi, pare, fillEsquerre(pare)); // sobreescr.
         else if ( (fillEsquerre(pare) == null) &&
                   (fillDret(pare) != null) )        // noms t fill dret
            reemplacarSubarbre(avi, pare, fillDret(pare));    // sobreescriu
         else {                                            // t dos fills
            Posicio<E> min = fillDret(pare);      // mnim valor branca dreta
            while (fillEsquerre(min)!=null)
            	min=fillEsquerre(min);
            intercanviar(pare, min);          // sobreescriu el node
            //  i esborra el que contenia el mnim del subarbre dret
            esborrarEnSubarbre(pare, fillDret(pare), min.getElem());
         }
      }
      equilibrar(avi);    // si cal, equilibra l'arbre o subarbre de nodes
      return  elemEsborrat;                                              // trobat
   }

   
   /**
    * Accessor de lectura d'un element de l'arbre. Si no el troba,
    * d'acord amb el comparador, retorna null.
    * @param pare  posici de referncia (inicialment l'arrel)
    * @param elemComp  element comparable que es vol consultar
    * @exception ExcepcioElementsNoComparables  si algun element s no
    *            comparable
    * @return  l'element interessat, o b null si no existeix
    */
   private E consultarEnSubarbre(Posicio<E> pare, E elemComp) throws ExcepcioElementsNoComparables {
   	  boolean trobat=false;
      int comp;
      while (pare!=null && !trobat) {
         comp = comparar(elemComp, pare.getElem());
         trobat = comp==0;
         if      (comp < 0)  pare = fillEsquerre(pare);
         else if (comp > 0)  pare = fillDret(pare);
      }
      return trobat ? pare.getElem() : null;
   }

   
   /**
    * Mtode per a saber si hi ha un comparador especfic. Si no hi s
    * s'utilitza l'operaci de comparaci de java.lang.Comparable.
    * @return  cert o fals, segons si hi ha o no hi ha un comparador
    *          especfic assignat
    */
   public boolean hiHaComparador() { return  (comparador != null); }

   
   /**
    * Mtode protegit que compara els elements rebuts. Si a la classe
    * principal hi ha definit un comparador, el mtode delega en el
    * mtode java.util.Comparator.compare(o1, o2); altrament delega en
    * el mtode java.lang.Comparable.compareTo(o) implementat per la
    * classe de l'element.
    * @param elem1  primer objecte de la mateixa classe
    * @param elem2  segon objecte de la mateixa classe
    * @exception ExcepcioElementsNoComparables  si algun element s no
    *            comparable
    * @return  un enter negatiu, zero o positiu, segons si el primer
    *          element t menys, igual o ms prioritat que el segon
    *          element, d'acord amb la implementaci del comparador
    * @pre elem1!=null && elem2!=null, ExcepcioElementsNoComparables
    */
   protected int comparar(E elem1,E elem2) throws  ExcepcioElementsNoComparables {
    int comp = 0;
    if (comparador == null)
       comp = ((Comparable<E>)elem1).compareTo(elem2);
    else
    	comp = comparador.compare(elem1, elem2);
    return  comp;
   }

   /**
    * Mtode que han d'implementar les subclasses que necessitin treballar
    * amb arbres equilibrats. L'eficincia temporal de les operacions
    * d'accs individual als elements d'un arbre de cerca depn
    * exclusivament de la seva alada i, en el pitjor dels casos, pot
    * esdevenir lineal. Hi ha variants d'arbres que garanteixen un cost
    * logartmic en base a reduir el nombre de nivells. Per a equilibrar
    * l'arbre han de fer un tractament dels nodes des del lloc d'inserci
    * o de supressi fins a l'arrel de l'arbre. Aquest mtode s cridat a
    * cada pas per afegeix() i per esborra(); si no est redefinit no fa
    * res. Per dna l'oportunitat a les classes descendents d'implementar
    * un tractament que garanteixi l'equilibri tot aprofitant el mateix
    * codi d'afegir i d'esborrar que han heretat.
    * @param abin  arbre o subarbre binari que s'ha d'equilibrar
    */
   protected void equilibrar(Posicio<E> abin) {}

}
