
package uoc.ei.tads;

/**
 * Classe que implementa les operacions de qualsevol arbre binari, el qual
 * es caracteritza per organitzar els seus elements (nodes) formant una
 * jerarquia: tot node (tret de l'arrel que es el cap de la jerarquia) s
 * descendent d'un node nic, i pot ser ascendent d'un mxim de dos nodes
 * (quan no t descendents s'anomena fulla).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class ArbreBinariEncadenatImpl<E>  extends ArbreBinari<E> {
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Punter a l'arrel de l'arbre. */
   protected NodeArbre<E> root;


   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() {
      return  (root == null) ? 0 : root.nombreNodes();
   }

   
   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  ( root == null ); }


   /**
    * Accessor de lectura de l'arrel de l'arbre, si n'hi ha.
    * @return  arrel de l'arbre, null en cas que l'arbre
    * estigui buit.
    */
   public Posicio<E> arrel()  throws  ExcepcioContenidorBuit {
      return  root;
   }


   /**
    * Afegeix un element com a nou fill de la posici rebuda, si es pot.
    * Si el pare s null, s'afegeix a l'arrel; si s una fulla, s'afegeix
    * com a fill esquerre; altrament, s'afegeix a la posici lliure.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @exception ExcepcioPosicioInvalida  si no cap cap ms fill
    * @return  nou fill; o l'arrel, si el pare era null
    * @pre pare==null || nombreFills(pare)<2, ExcepcioPosicioInvalida
    */
   public Posicio<E> afegir(Posicio<E> pare, E elem) {
      NodeArbre<E> nou = crearNode(pare,elem);
      if (pare == null)
		  root = nou;
      else {
         if (fillEsquerre(pare) == null)
         	setFillEsquerre(pare,nou);
         else
         	setFillDret(pare,nou);
      }
      return  nou;
   }

   
   /**
    * Afegeix un element com a fill esquerre de la posici rebuda, si es
    * pot.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @exception ExcepcioPosicioInvalida  si alguna posici s null o no
    *            vlida
    * @return  nou fill esquerre
    * @pre pare!=null && ((NodeArbre)pare).getFillEsquerre() == null, ExcepcioPosicioInvalida
    */
   public Posicio<E> afegirFillEsquerre(Posicio<E> pare, E elem) {
      NodeArbre<E> nou = crearNode(pare,elem);
      setFillEsquerre(pare,nou);
      return  nou;
   }

   
   /**
    * Afegeix un element com a fill dret de la posici rebuda, si es pot.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @exception ExcepcioPosicioInvalida  si alguna posici s null o no
    *            vlida
    * @return  nou fill dret
    * @pre pare!=null && ((NodeArbre)pare).getFillDret() == null, ExcepcioPosicioInvalida
    */
   public Posicio<E> afegirFillDret(Posicio<E> pare, E elem) {
      NodeArbre<E> nou = crearNode(pare,elem);
      setFillDret(pare,nou);
      return  nou;
   }

   
   /**
    * Accessor de lectura del fill esquerre d'una posici de l'arbre.
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  fill esquerre de la posici rebuda; o null si no en t
    * @pre !estaBuit() && node!=null, ExcepcioPosicioInvalida
    */
   public Posicio<E> fillEsquerre(Posicio<E> node) {
      return  ((NodeArbre<E>)node).getFillEsquerre();
   }

   
   /**
    * Accessor de lectura del fill dret d'una posici de l'arbre.
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  fill dret de la posici rebuda; o null si no en t
    * @pre !estaBuit() && node!=null, ExcepcioPosicioInvalida
    */
   public Posicio<E> fillDret(Posicio<E> node) {
      return  ((NodeArbre<E>)node).getFillDret();
   }

   
   /**
    * Esborra el subarbre representat per la posici fill, si es pot.
    * Si la posici del pare s null, esborra l'arbre sencer.
    * @param pare  posici del pare; pot ser null
    * @param fill  posici del fill
    * @exception ExcepcioPosicioInvalida  si alguna posici s no vlida
    * @pre fill!=null, ExcepcioPosicioInvalida
    * @pre pare==null || fillEsquerre(pare)==fill || fillDret(pare)==fill, ExcepcioPosicioInvalida
    */
   public void esborrar(Posicio<E> pare, Posicio<E> fill) {
      if (pare == null)
      	root = null;
      else {
         if (fillEsquerre(pare) == fill)
         	setFillEsquerre(pare,null);
         else
         	setFillDret(pare,null);
      }
   }

   
   /**
    * Reemplaa l'element contingut a la posici rebuda.
    * @param elem  nou element
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  element que hi havia a la posici
    * @pre node!=null, ExcepcioPosicioInvalida
    */
   public E reemplacar(Posicio<E> node, E elem) {
      E old = node.getElem();
      ((NodeArbre<E>)node).setElem(elem);
      return  old;
   }

   
   /**
    * Intercanvia els elements continguts a les posicions rebudes.
    * Les posicions no s'intercanvien, sin que es modifica el
    * seu contingut.
    * @param node1  primera de les dues posicions de referncia
    * @param node2  segona de les dues posicions de referncia
    * @exception ExcepcioPosicioInvalida  si alguna posici s null o no
    *            vlida
    * @pre node1!=null && node2!=null, ExcepcioPosicioInvalida
    */
   public void intercanviar(Posicio<E> node1, Posicio<E> node2) {
      E tmp = node1.getElem();
      ((NodeArbre<E>)node1).setElem(node2.getElem());
      ((NodeArbre<E>)node2).setElem(tmp);
   }

   
   /**
    * Reemplaa el subarbre representat per la posici fill, si es pot.
    * Si la posici del pare s null, substitueix l'arbre sencer. Aquest
    * mtode auxiliar noms s accessible dins del package i s utilitzat
    * per a esborrar un node amb menys de dos fills.
    * @param pare  posici del pare; pot ser null
    * @param fill  posici del fill
    * @param nou  posici amb el nou arbre o subarbre; pot ser null
    * @exception ExcepcioPosicioInvalida  si alguna posici s no vlida
    * @pre fill!=null, ExcepcioPosicioInvalida
    * @pre pare==null && fill==root, ExcepcioPosicioInvalida
    * @pre pare!=null && (fill==fillEsquerre(pare) || fill==fillDret(pare)), ExcepcioPosicioInvalida
    */
   
   protected void reemplacarSubarbre(Posicio<E> pare, Posicio<E> fill, Posicio<E> nou) {
      if (pare == null)  root = (NodeArbre<E>)nou;              // substitueix tot l'arbre
      else {
      	if (fillEsquerre(pare)==fill)
      		setFillEsquerre(pare,(NodeArbre<E>)nou);
      	else
      		setFillDret(pare,(NodeArbre<E>)nou);
      }
   }


   /**
    * Crea un nou node, amb dos encadenaments a node, que emmagatzema un
    * element. El mtode ha de ser sobreescrit per les subclasses que necessitin
    * altres tipus de node.
    * @param elem  element que es vol desar al node
    * @return  node arbre binari nou que emmagatzema l'element
    */
   protected NodeArbre<E> crearNode(Posicio<E> pare, E elem) {
      return  new NodeArbre<E>(elem);
   }

   
   private void setFillEsquerre(Posicio<E> node,NodeArbre<E> fill) {
      ((NodeArbre<E>)node).setFillEsquerre(fill);
   }

   
   private void setFillDret(Posicio<E> node,NodeArbre<E> fill) {
      ((NodeArbre<E>)node).setFillDret(fill);
   }

   
   /**
    * Classe que implementa un node amb dos encadenaments a node.
    * Noms facilita les operacions bsiques: constructors i accessors de
    * lectura i d'escriptura. Pot emmagatzemar qualsevol element (Object).
    */
   protected static class NodeArbre<EN>  implements Posicio<EN> {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
	   * serialitzables de la mateixa classe. L'deitificador es calcula
	   * mitjanant un mtode de la classe Utilitats. */
	  private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	 	
      /** Element contingut al node. */
      protected EN element;

      /** Encadenament al fill esquerre. */
      protected NodeArbre<EN> fillEsquerre;

      /** Encadenament al fill dret. */
      protected NodeArbre<EN> fillDret;

      /** Constructor sense parmetres. Dna valor nul als atributs. */
      public NodeArbre() { super(); }

      /**
       * Constructor amb un parmetre. Assigna el valor rebut a l'element
       * del node pare i dna valor null a les posicions filles.
       * @param elem  valor de l'element que ha d'anar al node pare
       */
      public NodeArbre(EN elem) { element = elem; }

      /**
       * Constructor amb tres parmetres.
       * @param fe  fill esquerre
       * @param elem  valor de l'element contingut al node pare
       * @param fd  fill dret
       */
      public NodeArbre(NodeArbre<EN> fe, EN elem, NodeArbre<EN> fd) {
         setElem(elem);  fillEsquerre = fe;  fillDret = fd;
      }

      /**
       * Compta recursivament els nodes que cont un arbre o subarbre.
       * @return  zero si est buit; altrament retorna el nombre de nodes
       * del subarbre esquerre ms el del subarbre dret ms u (arrel)
       */
      public int nombreNodes() {
      	int nnodes=1;
      	NodeArbre<EN> fe=getFillEsquerre();
      	NodeArbre<EN> fd=getFillDret();
      	if (fe!=null) nnodes+=fe.nombreNodes();
      	if (fd!=null) nnodes+=fd.nombreNodes();
      	return nnodes;
      }

      /**
       * Accessor d'escriptura del valor contingut al node.
       * @param elem  nou valor de l'element contingut al node
       */
      public void setElem(EN elem) { element = elem; }

      /**
       * Accessor de lectura de l'element contingut al node.
       * @return  element contingut al node
       */
      public EN getElem() { return  element; }

      /**
       * Accessor d'escriptura del fill esquerre.
       * @param fe  valor per al fill esquerre
       */
      public void setFillEsquerre(NodeArbre<EN> fe) { fillEsquerre = fe; }

      /**
       * Accesor de lectura del fill esquerre.
       * @return  fill esquerre
       */
      public NodeArbre<EN> getFillEsquerre() { return  fillEsquerre; }

      /**
       * Accessor d'escriptura del fill dret.
       * @param fd  valor per al fill dret
       */
      public void setFillDret(NodeArbre<EN> fd) { fillDret = fd; }

      /**
       * Accesor de lectura del fill dret.
       * @return  fill dret
       */
      public NodeArbre<EN> getFillDret() { return  fillDret; }

      /**
       * Mtode que redefineix la conversi de l'objecte a String per
       * facilitar la depuraci del codi. Delega en el mtode toString()
       * de l'element emmagatzemat a la posici.
       * @return  cadena de carcters representativa de l'element
       */
      public String toString() {
         return  (element == null) ? "null" : element.toString();
      }
   }

}
