package uoc.ei.tads;
/**
 * Classe que implementa les operacions de qualsevol arbre binari, el qual
 * es caracteritza per organitzar els seus elements (nodes) formant una
 * jerarquia: tot node (tret de l'arrel que es el cap de la jerarquia) s
 * descendent d'un node nic, i pot ser ascendent d'un mxim de dos nodes
 * (quan no t descendents s'anomena fulla).
 *
 * En la representaci seqencial de l'arbre s'organitzen els nodes dins
 * d'un vector de manera que els fills i el pare sn accessibles aplicant
 * noms una frmula. La primera posici del vector contindr sempre
 * l'arrel de l'arbre; si un node est a la posici [k], el seu fill
 * esquerre estar a [2*k] i el seu fill dret a [2*k+1].
 *
 * Aquesta representaci t dues operacions modificadores (afegir al final
 * i esborrar l'ltima posici) que permeten mantenir una estructura
 * d'arbre quasicomplet; s a dir, no hi haur cap posici lliure en un
 * recorregut per nivells de l'arbre. Ats que la primera posici [0] del
 * vector no s'utilitza per facilitar els clculs esmentats, si el nombre
 * de caselles del vector s potncia de dos, l'arbre podr arribar a ser
 * complet.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class ArbreBinariVectorImpl<E>  extends ArbreBinari<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


   /** Capacitat mxima, per defecte, del contenidor. */
   public static final int MAXIM_ELEMENTS_PER_DEFECTE = 256;

   /** Nombre d'elements que hi ha actualment al contenidor. */
   protected int n;

   /** Taula d'elements del contenidor. Les posicions comencen pel zero.*/
   protected Rang<E>[] elems;

   /** Constructor sense parmetres (capacitat mxima, per defecte). */
   public ArbreBinariVectorImpl() {
   	  this(MAXIM_ELEMENTS_PER_DEFECTE);
   }

   /**
    * Constructor amb un parmetre. Crea un vector amb una capacitat de
    * mxim.
    * @exception ExcepcioParametreIncorrecte  si la capacitat mxima del
    *                                         nou arbre s negativa
    * @pre max>=0 , new ExcepcioParametreIncorrecte("el nombre d'elements mxim no pot ser negatiu")
    * @param max  nombre mxim d'elements de l'arbre
    */
   public ArbreBinariVectorImpl(int max) {
      elems = new Rang[max];
      n = 0;
   }

   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  n; }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  ( n == 0 ); }

   /**
    * Mtode per a comprovar si el contenidor est ple.
    * @return  cert o fals, segons si est ple o no ho est
    */
   public boolean estaPle() { return  ( n == elems.length ); }

   /**
    * Accessor de lectura de l'arrel de l'arbre, si n'hi ha.
    * @pre !estaBuit(), new ExcepcioContenidorBuit("No es pot accedir a l'arrel d'un arbre buit")
    * @exception ExcepcioContenidorBuit  si l'arbre est buit
    * @return  arrel de l'arbre
    */
   public Posicio<E> arrel() { return  elems[0]; }


   /**
    * Afegeix un element com a nou fill de la posici rebuda.
    * Si el pare s null, s'afegeix a l'arrel; si s una fulla, s'afegeix
    * com a fill esquerre; i si el pare t nicament un fill, s'afegeix com a
    * l'altre fill.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @pre posicioValida(pare) &&
    * 	    (indexFillEsquerre(pare)<maxim && !teFillEsquerre(pare)) ||
    * 	    (indexFillDret(pare)<maxim && !teFillDret(pare)), ExcepcioPosicioInvalida
    * @exception ExcepcioPosicioInvalida  Si la posici de pare s invlida per a poder
    * afegir un fill. Aix inclou tamb els casos en que el pare ja tingui dos fills i
    * el cas en que els fills caiguin fora del rang mxim perms al crear l'arbre.
    * @return La posici del node afegit.
    */
   public Posicio<E> afegir(Posicio<E> pare, E elem)
   {
   	  Posicio<E> posicioNouFill;
   	  if (fillEsquerre(pare)!=null)
   	  	 posicioNouFill=afegirFillDret(pare,elem);
   	  else
   	  	 posicioNouFill=afegirFillEsquerre(pare,elem);
   	  return posicioNouFill;
   }


   /**
    * Afegeix un element com a fill esquerre de la posici rebuda.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @pre posicioValida(pare) &&
    * 	    indexFillEsquerre(pare)<maxim && !teFillEsquerre(pare)
    * @exception ExcepcioPosicioInvalida  Si la posici de pare s invlida per a poder
    * afegir un fill. Aix inclou tamb els casos en que el pare ja tingui fill esquerre
    * i el cas en que aquest caigui fora del rang mxim perms al crear l'arbre.
    * @return La posici del node afegit.
    */
   public Posicio<E> afegirFillEsquerre(Posicio<E> pare, E elem)
   {
   	  int index = getIndexFillEsquerre(pare);
   	  n++;
   	  elems[index]=new Rang<E>(index,elem);
   	  return elems[index];
   }


   /**
    * Afegeix un element com a fill dret de la posici rebuda, si es pot.
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @pre posicioValida(pare) &&
    * 	    indexFillDret(pare)<maxim && !teFillDret(pare)
    * @exception ExcepcioPosicioInvalida  Si la posici de pare s invlida per a poder
    * afegir un fill. Aix inclou tamb els casos en que el pare ja tingui fill dret
    * i el cas en que aquest caigui fora del rang mxim perms al crear l'arbre.
    * @return La posici del node afegit.
    */
   public Posicio<E> afegirFillDret(Posicio<E> pare, E elem)
   {
 	  int index = getIndexFillDret(pare);
   	  n++;
   	  elems[index]=new Rang<E>(index,elem);
   	  return elems[index];
   }


   /**
    * Accessor de lectura del pare d'una posici de l'arbre, si n'hi ha.
    * @param pos  posici de referncia
    * @return  pare de la posici rebuda; o null si era l'arrel.
    */
   public Posicio<E> pare(Posicio<E> pos)
   {
   	  if (getIndex(pos)==0) return null;
   	  return elems[getIndexPare(pos)];
   }

   /**
    * Accessor de lectura del fill esquerre d'una posici de l'arbre.
    * @param pos  posici de referncia
    * @pre posicioValida(pos)
    * @exception ExcepcioPosicioInvalida  si la posici no s vlida.
    * @return  fill esquerre de la posici rebuda; o null si no en t.
    */
   public Posicio<E> fillEsquerre(Posicio<E> pos)
   {
 	  int index=getIndexFillEsquerre((Rang<E>)pos);
   	  if (index>=elems.length || elems[index]==null)
   	  	 return null;
   	  return elems[index];
   }

   /**
    * Accessor de lectura del fill dret d'una posici de l'arbre.
    * @param pos  posici de referncia
    * @pre posicioValida(pos)
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  fill dret de la posici rebuda; o null si no en t.
    */
   public Posicio<E> fillDret(Posicio<E> pos)
   {
	  int index=getIndexFillDret((Rang<E>)pos);
   	  if (index>=elems.length || elems[index]==null)
   	  	 return null;
   	  return elems[index];
   }

   /**
    * Esborra el subarbre representat per la posici fill, si es pot.
    * Si la posici del pare s null, esborra l'arbre sencer.
    * @see  ArbreBinariVectorImpl#esborrar(Posicio, Posicio)
    * @param pare  posici del pare; pot ser null
    * @param fill  posici del fill
    * @pre esPareDe(pare,fill)
    * @exception ExcepcioPosicioInvalida  si alguna posici s no vlida o
    * b pare no s el pare de fill.
    */
   public void esborrar(Posicio<E> pare, Posicio<E> fill)
   {
 	  Posicio<E> netEsquerre = fillEsquerre(fill);
   	  Posicio<E> netDret = fillDret(fill);
   	  if (netEsquerre!=null)
   	  	  esborrar(fill,netEsquerre);
   	  if (netDret!=null)
   	  	  esborrar(fill,netDret);
   	  int indexFill=getIndex(fill);
   	  n--;
   	  elems[indexFill]=null;
   }


   /**
    * Reemplaa l'element contingut a la posici rebuda.
    * @param elem  nou element
    * @param pos  posici de referncia
    * @pre posicioValida(pos)
    * @return  element que hi havia a la posici abans de reemplaar-lo.
    */
   public E reemplacar(Posicio<E> pos, E elem)
   {
   	  Rang<E> rang=(Rang<E>)pos;
   	  E old=rang.getElem();
   	  rang.setElem(elem);
   	  return old;
   }

   /**
    * Intercanvia dues posicions dins de l'arbre. Els objectes
    * posici s'intercanvien de posici, amb lo qual el seu
    * contingut es mant constant.
    * @param pos1  primera de les dues posicions de referncia.
    * @param pos2  segona de les dues posicions de referncia.
    * @pre posicioValida(pos1) && posicioValida(pos2)
    */
   public void intercanviar(Posicio pos1, Posicio pos2)
   {
 	  Rang<E> r1=(Rang<E>)pos1;
   	  Rang<E> r2=(Rang<E>)pos2;
   	  int index1=r1.getIndex();
   	  int index2=r2.getIndex();
   	  r1.setIndex(index2);
   	  r2.setIndex(index1);
   	  elems[index1]=r2;
   	  elems[index2]=r1;
   }


   private int getIndex(Posicio<E> pos) {
   	  if (pos==null) return -1;
	  return ((Rang)pos).getIndex();
   }


   private int getIndexPare(Posicio<E> pos) {
   	  int index=((Rang)pos).getIndex();
   	  if (index==0) return -1;
	  return (index-1)/2;
   }


   private int getIndexFillEsquerre(Posicio<E> pos) {
 	  if (pos==null) return 0;
	  return ((Rang)pos).getIndex()*2+1;
   }


   private int getIndexFillDret(Posicio<E> pos) {
 	  if (pos==null) return 0;
 	  return ((Rang)pos).getIndex()*2+2;
 }


}
