package uoc.ei.tads;
/**
 * Els arbres sn estructures que relacionen els seus elements, anomenats
 * nodes, formant jerarquies: tot node (tret de l'arrel que es el cap de
 * la jerarquia) s descendent d'un node nic, i pot ser ascendent
 * d'altres nodes (quan no t descendents s'anomena fulla). Quan un node
 * pot tenir un nombre indeterminat de fills parlem d'arbres generals
 * (general tree) i, si en t un nombre fix N, d'arbres d'ordre N (n-ary
 * tree); en aquests ltims destaca el cas de N = 2, els anomenats arbres
 * binaris (binary tree).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class ArbreGeneralDelegImpl<E>  extends ArbreAbstracte<E> {
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Arbre binari que implementa les operacions per delegaci. */
   protected ArbreBinariEncadenatImpl<E> ab;
   
   
   public ArbreGeneralDelegImpl() {
	   ab = creaArbreRepresentacio();
   }
   
   
   protected ArbreBinariEncadenatImpl<E> creaArbreRepresentacio() {
	   return new ArbreBinariEncadenatImpl<E>();
   }


   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  ab.nombreElems(); }


   /**
    * Accessor de lectura de l'arrel de l'arbre, si n'hi ha.
    * @return  arrel de l'arbre
    */
   public Posicio<E> arrel() { return  ab.arrel(); }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions filles
    * de la posici de referncia, simultanis i independents entre ells.
    * @param pare  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  enumeraci de les posicions filles
    */
   public Recorregut<E> fills(Posicio<E> pare)
   {
      if (estaBuit() || (pare == null))
         throw  new ExcepcioPosicioInvalida();

      return new RecorregutFills<E>(this,pare);
   }

   /**
    * Afegeix un element com a nou fill de la posici rebuda, si es pot.
    * Si el pare s null, s'afegeix a l'arrel; si s una fulla, s'afegeix
    * com a primer fill (fill ms a l'esquerra);  altrament, s'afegeix a
    * la primera posici lliure (d'esquerra a dreta).
    * @param pare  posici de referncia
    * @param elem  element que es vol afegir a l'arbre
    * @return  nou fill; o l'arrel, si el pare era null
    */
   public Posicio<E> afegir(Posicio<E> pare, E elem)
   {
      if ( (pare == null) || esFulla(pare) )
         return  ab.afegir(pare, elem);

      pare = ab.fillEsquerre(pare);
      while (ab.fillDret(pare) != null)  pare = ab.fillDret(pare);

      return  ab.afegirFillDret(pare, elem);
   }

   /**
    * Reemplaa l'element contingut a la posici rebuda.
    * @param elem  nou element
    * @param pos  posici de referncia
    * @return  element que hi havia a la posici
    */
   public E reemplacar(Posicio<E> pos, E elem)
   {
      return  ab.reemplacar(pos, elem);
   }

   /**
    * Intercanvia els elements continguts a les posicions rebudes.
    * Les posicions no sn intercanviades.
    * @param pos1  primera de les dues posicions de referncia
    * @param pos2  segona de les dues posicions de referncia
    */
   public void intercanviar(Posicio<E> pos1, Posicio<E> pos2)
   {
      ab.intercanviar(pos1, pos2);
   }

   /**
    * Esborra el subarbre representat per la posici fill, si es pot.
    * Si la posici del pare s null, esborra l'arbre sencer.
    * @param pare  posici del pare; pot ser null
    * @param fill  posici del fill
    * @exception ExcepcioPosicioInvalida  si alguna posici s no vlida
    */
   public void esborrar(Posicio<E> pare, Posicio<E> fill)
               throws  ExcepcioPosicioInvalida
   {
      if (estaBuit() || (fill == null))
         throw  new ExcepcioPosicioInvalida();

      if (pare == null)  ab.esborrar(pare, fill);  // esborra arbre sencer
      else
      {
         if (fill != ab.fillEsquerre(pare))        // segon o altres fills
         {
            for ( pare = ab.fillEsquerre(pare);
                  ab.fillDret(pare) != fill;
                  pare = ab.fillDret(pare) );          // recorregut fills
         }
         ab.reemplacarSubarbre(pare, fill, ab.fillDret(fill)); // sobreescriu fill
      }
   }

   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutFills<E> implements Recorregut<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();

      
      /** Arbre que s'est recorrent. */
      protected ArbreGeneralDelegImpl<E> arbre;

      
      /** Posici auxiliar. */
      protected Posicio<E> fill;


      /**
       * Constructor amb un parmetre.
       * @param pare  posicio de referncia; s'espera que sigui una
       *              posici vlida
       */
      public RecorregutFills(ArbreGeneralDelegImpl<E> arbre,Posicio<E> pare) {
      	 this.arbre=arbre;
         fill = arbre.ab.fillEsquerre(pare);
      }

      
      /**
       * Comprova si hi ha una primera o segent posici. s sensible a
       * eventuals alteracions de l'estructura de posicions. Retorna fals
       * si est buit o ja s'ha visitat l'ltima posici.
       * @return  cert o fals, segons si es pot avanar o no es pot
       */
      public boolean hiHaSeguent() { return  fill != null; }

      
      /**
       * Primer avana, si es pot, i desprs retorna la posici.
       * Si no hi ha segent posici llana una excepci.
       * @return  segent posici
       */
      public Posicio<E> seguent()  throws  ExcepcioPosicioInvalida {
          if ( ! hiHaSeguent() )  throw new ExcepcioPosicioInvalida(
                                           "No hi ha seguent");

         Posicio<E> temp = fill;
         fill = arbre.ab.fillDret(fill);
         return  temp;
      }
   }
}
