package uoc.ei.tads;

/**
 * Classe que implementa les operacions d'un conjunt per delegaci en un
 * arbre binari AVL.
 *
 * Els conjunts sn estructures que emmagatzemen elements no repetits. La
 * classe dels objectes ha de disposar d'una operaci d'igualtat.
 *
 * En aquesta implementaci s'espera que la classe dels elements
 * implementi la interfcie java.lang.Comparable o b que es faciliti un
 * java.util.Comparator com a parmetre del constructor.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */

public class ConjuntAVLImpl<E> extends ConjuntAbstracte<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Arbre binari AVL que implementa les operacions per delegaci. */
   protected ArbreAVL<E> avl;

   
   /**
    * Constructor sense parmetres. S'espera que les classes dels elements
    * implementin la interfcie java.lang.Comparable.
    */
   public ConjuntAVLImpl() { avl = new ArbreAVL<E>(); }

   
   /**
    * Constructor amb un parmetre i elements d'una classe comparable amb
    * el comparador donat.
    * @param comparador  comparador que permet deduir la prioritat
    */
   public ConjuntAVLImpl(java.util.Comparator<E> comparador)
   {
      avl = new ArbreAVL<E>(comparador);
   }

   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  avl.nombreElems(); }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  avl.estaBuit(); }

   /**
    * Afegeix un element, si es pot.
    * @param elem  element que es vol afegir al conjunt
    */
   public void afegir(E elem)
   {
   	avl.afegir(elem);
   }

   /**
    * Comprova si hi ha un element.
    * @param elem  element de referncia
    * @return  cert o fals, segons si troba o no troba l'element
    */
   public boolean hiEs(E elem)
   {
      return  avl.consultar(elem) != null;
   }

   /**
    * Esborra un element, si es pot.
    * @param elem  element de referncia
    * @return  element esborrat; o null, si no hi era
    */
   public E esborrar(E elem)
   {
   		return avl.esborrar(elem);
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements associats amb les claus
    */
   public Iterador<E> elements() { return  avl.elements(); }

   /**
    * Mtode que sobreescriu Object.toString(). Treu els elements separats
    * pel salt de lnia de la plataforma.
    * @return  llistat dels elements en un recorregut inordre
    */
   public String toString()
   {
 	  return Utilitats.toStringContenidorDelegacio("ConjuntAVLImpl",avl);
   }
}
