package uoc.ei.tads;
/**
 * Classe que implementa les operacions bsiques d'un conjunt per
 * delegaci en una taula de dispersi.
 *
 * Els conjunts sn estructures que emmagatzemen elements no repetits. La
 * classe dels objectes ha de disposar d'una operaci d'igualtat.
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */

public class ConjuntTaulaImpl<E>  extends ConjuntAbstracte<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


   /** Taula de dispersi que implementa les operacions per delegaci. */
   protected TaulaDispersio<E,E> td;


   /** Constructor sense parmetres (mida de la taula per defecte). */
   public ConjuntTaulaImpl()
   {
   	td = new TaulaDispersio<E,E>();
   }


   /** Constructor amb un parmetre (taula de la mida donada).
    * @param mida  mida de la taula de dispersi
    * @exception ExcepcioParametreIncorrecte  si la mida s negativa
    * @pre mida>=0, new ExcepcioParametreIncorrecte("mida negativa")
    */
   public ConjuntTaulaImpl(int mida)
   {
      td = new TaulaDispersio<E,E>(mida);
   }


   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  td.nombreElems(); }


   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  td.estaBuit(); }


   /**
    * Afegeix un element, si es pot. Utilitza el mateix element com a clau
    * i com a valor. S'implementa per compatibilitat amb la interfcie.
    * Es recomana utilitzar afegir(clau, valor).
    * @param elem  element que es vol afegir al conjunt
    */
   public void afegir(E elem)
   {
   	td.afegir(elem, elem);
   }


   /**
    * Comprova si hi ha un element amb una determinada clau.
    * @param elem  clau associada a un element
    * @exception ExcepcioContenidorBuit  si la taula est buida
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @return  cert o fals, segons si troba o no troba la clau
    */
   public boolean hiEs(E elem)
   {
      return  td.consultar(elem) != null;
   }


   /**
    * Esborra la clau i l'element associat, si es pot.
    * @see  ClauValor
    * @param clau  clau de referncia
    * @exception ExcepcioContenidorBuit  si la taula est buida
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @return  element esborrat associat amb la clau;
    *          o null, si no hi era
    */
   public E esborrar(E clau)
   {
   	return  td.esborrar(clau);
   }


   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements associats amb les claus
    */
   public Iterador<E> elements() { return  td.elements(); }


   /**
    * Mtode que sobreescriu Object.toString(). Treu parelles d'elements.
    * Separa una parella de la segent amb el salt de lnia de la
    * plataforma.
    * @return  llistat de claus (entre claudtors) seguides de l'element
    *          associat a cada clau
    */
   public String toString()
   {
	  return Utilitats.toStringContenidorDelegacio("ConjuntTaulaImpl",td);
   }
}
