package uoc.ei.tads;
/**
 * Classe que implementa les operacions de les cues prioritries (priority
 * queue) i es distingeix perqu els elements no s'insereixen pel final,
 * sin que s'ordenen segons una prioritat que tenen associada.
 * Les operacions de supressi i consulta afecten sempre l'element que t
 * una prioritat ms petita. En conseqncia, els elements hauran de
 * permetre una operaci de comparaci d'on es pugui deduir la prioritat.
 *
 * En la representaci seqencial de l'arbre mitjanant un pil
 * s'organitzen els nodes de l'arbre dins d'un vector de manera que els
 * fills i el pare sn accessibles aplicant noms una frmula. La primera
 * posici del vector contindr sempre l'arrel de l'arbre que s,
 * precisament, l'element ms petit; si un node est a la posici [k], el
 * seu fill esquerre estar a [2*k] i el seu fill dret a [2*k+1].
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class CuaAmbPrioritat<E>  implements Cua<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


	/** Classe que exten el comportament d'un arbre binari amb dos mtodes que ens
	 * proporcionen la funcionalitat d'un arbre quasicomplet.
	 * @param <E>
	 */

	protected static class ArbreBinariQuasicompletVectorImpl<E> extends ArbreBinariVectorImpl<E> {

		   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
		    * serialitzables de la mateixa classe. L'deitificador es calcula
		    * mitjanant un mtode de la classe Utilitats. */
		   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


		   public ArbreBinariQuasicompletVectorImpl() { super(); }
		   public ArbreBinariQuasicompletVectorImpl(int max) { super(max); }


		   /** Retorna el darrer element en un recorregut per nivells de l'arbre quasicomplet.
		    * @pre !estaBuit(), ExcepcioContenidorBuit
		    * @return El darrer element en un recorregut per nivells.
		    */
		   public Posicio<E> darrer() {
		   	  return elems[n-1];
		   }

		   /** Afegeix un element a continuaci de la darrera posici de l'arbre en
		    * un recorregut per nivells. Mant la seva condici d'arbre quasicomplet.
		    * @param elem  element que es vol afegir a l'arbre
		    * @pre !estaPle(), ExcepcioContenidorPle
		    * @return  nova ltima posici
		    */
		   public Posicio<E> afegirDarrer(E elem) {
		   	  n++;
		      elems[n-1] = new Rang<E>(n-1, elem);
		      return darrer();
		   }

		   /**
		    * Esborra la darrera posici ocupada de l'arbre en un recorregut per
		    * nivells. Mant la seva condici d'arbre quasicomplet.
		    * @pre !estaBuit(), ExcepcioContenidorBuit
		    * @return L'element esborrat
		    */
		   public Posicio<E> esborrarDarrer() {
		   	  n--;
		   	  Posicio<E> aux=elems[n];
		      elems[n] = null;
		      return  aux;
		   }

	}



   /** Arbre binari vector que serveix de suport a la cua prioritria. */
   protected ArbreBinariQuasicompletVectorImpl<E> heap;

   /**
    * Comparador concret que permet deduir la prioritat entre els
    * elements. Pot tenir valor null i aleshores s'utilitza la interfcie
    * java.lang.Comparable
    */
   protected java.util.Comparator<E> comparador;

   /**
    * Constructor sense parmetres (capacitat mxima, per defecte) i
    * elements d'una classe que implementa java.lang.Comparable.
    */
   public CuaAmbPrioritat() {
   	  heap = new ArbreBinariQuasicompletVectorImpl<E>();
   }

   /**
    * Constructor amb un parmetre (capacitat donada) i elements d'una
    * classe que implementa java.lang.Comparable.
    * @param max  nombre mxim d'elements que pot contenir la cua
    */
   public CuaAmbPrioritat(int max) {
      heap = new ArbreBinariQuasicompletVectorImpl<E>(max);
   }

   /**
    * Constructor amb un parmetre (capacitat mxima, per defecte) i
    * elements d'una classe comparable amb el comparador donat.
    * @param comparador  comparador que permet deduir la prioritat
    */
   public CuaAmbPrioritat(java.util.Comparator<E> comparador) {
      this.comparador = comparador;
      heap = new ArbreBinariQuasicompletVectorImpl<E>();
   }

   /**
    * Constructor amb dos parmetres (capacitat donada) i elements d'una
    * classe comparable amb el comparador donat.
    * @param max  nombre mxim d'elements que pot contenir la cua
    * @param comparador  comparador que permet deduir la prioritat
    */
   public CuaAmbPrioritat(int max, java.util.Comparator<E> comparador) {
      this.comparador = comparador;
      heap = new ArbreBinariQuasicompletVectorImpl<E>(max);
   }

   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  heap.nombreElems(); }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  heap.estaBuit(); }

   /**
    * Mtode per a comprovar si el contenidor est ple.
    * @return  cert o fals, segons si est ple o no ho est
    */
   public boolean estaPle() { return  heap.estaPle(); }

   /**
    * Afegeix un element a la posici que li correspon, si hi cap.
    * @pre !heap.estaPle(), ExcepcioContenidorPle
    * @param elem  element comparable que es vol afegir a la cua
    */
   public void encuar(E elem) {
   	  pujarElement(novaDarreraPosicio(elem));
   }


   protected Posicio<E> novaDarreraPosicio(E elem) {
	   return heap.afegirDarrer(elem);
   }


   protected void pujarElement(Posicio<E> novaPosicio) {
   	  boolean trobat = false;
   	  while (!trobat) {
   	  	 Posicio<E> pare=heap.pare(novaPosicio);
   	  	 trobat= pare==null || comparar(novaPosicio,pare)>=0;
   	  	 if (!trobat)
   	  	 	heap.intercanviar(novaPosicio,pare);
   	  }
   }


   /**
    * Esborra l'element de menor prioritat, si n'hi ha algun.
    * @see  CuaAmbPrioritat#triarFillAIntercanviar(Posicio)
    * @pre !heap.estaBuit(), ExcepcioContenidorBuit
    * @return  primer element, que s el de menor prioritat
    */
   public E desencuar() {
	   if (estaBuit())
		   throw new ExcepcioContenidorBuit();
 	  E primer = heap.arrel().getElem();
 	  heap.intercanviar(heap.arrel(),heap.darrer());
 	  esborrarDarreraPosicio();
 	  enfonsarElement(heap.arrel());
 	  return primer;
   }


   protected Posicio<E> esborrarDarreraPosicio() {
	   return heap.esborrarDarrer();
   }


   protected void enfonsarElement(Posicio<E> posicioAOrdenar) {
 	  boolean trobat=false;
 	  while (!trobat) {
  	  	 Posicio<E> fillAIntercanviar=triarFillAIntercanviar(posicioAOrdenar);
  	  	 trobat = fillAIntercanviar==null || comparar(posicioAOrdenar,fillAIntercanviar)<=0;
  	  	 if (!trobat)
   	  	 	heap.intercanviar(posicioAOrdenar,fillAIntercanviar);
 	  }
   }


   /**
    * Accessor de lectura de l'element de menor prioritat, si n'hi ha.
    * Si es vol desencuar l'objecte s'ha d'esborrar seguidament.
    * @pre !estaBuit(), ExcepcioContenidorBuit
    * @exception ExcepcioContenidorBuit  si la cua est buida
    * @return  primer element, que s el de menor prioritat
    */
   public E primer() {
	   if (estaBuit())
		   throw new ExcepcioContenidorBuit();
	   return  heap.arrel().getElem();
   }


   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor inordre
    */
   public Iterador<E> elements() { return  heap.elements(); }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @return  enumeraci de les posicions del contenidor per nivell
    */
   public Recorregut posicions() { return  heap.posicions(); }

   /**
    * Mtode que sobreescriu Object.toString(). Treu els elements separats
    * pel salt de lnia de la plataforma.
    * @return  llistat dels elements en un recorregut per nivells
    */
   public String toString() { return  heap.toString(); }


   /**
    * Mtode protegit que compara els elements continguts a les posicions
    * rebudes. Si el constructor no ha definit un comparador, s'utilitza
    * java.lang.Comparable per deduir la prioritat entre ambds elements.
    * @exception ExcepcioPosicioInvalida  si alguna posici s null o no
    *            vlida
    * @exception ExcepcioElementsNoComparables  si algun dels elements
    *            continguts a les posicions s no comparable
    * @param pos1  primera de les dues posicions de referncia
    * @param pos2  segona de les dues posicions de referncia
    * @pre posicioValida(pos1) && posicioValida(pos2)
    * @return  un enter negatiu, zero o positiu, segons si l'element de la
    *          primera posici t menys, igual o ms prioritat que
    *          l'element de la segona posici, d'acord amb la
    *          implementaci del comparador
    */
   protected int comparar(Posicio<E> pos1, Posicio<E> pos2) {
      E elem1 = pos1.getElem();
      E elem2 = pos2.getElem();
      int comp = 0;
      if (comparador == null)
         comp = ((Comparable<E>)elem1).compareTo(elem2);
      else
      	comp = comparador.compare(elem1, elem2);
      return  comp;
   }


   /** Mtode auxiliar que donada una posici de l'arbre que no est ordenada dins l'arbre i
    * que volem ordenar, determina per quin dels seus dos fills hauria de ser intercanviada
    * (cas de produr-se quest intercanvi.
    * @param pos
    * @return
    */
	private Posicio<E> triarFillAIntercanviar(Posicio<E> pos) {
	   Posicio<E> fillDret=heap.fillDret(pos);
	   Posicio<E> fillEsquerre=heap.fillEsquerre(pos);
	   if (fillDret==null) return fillEsquerre;
	   if (fillEsquerre==null) return fillDret;
	   return comparar(fillEsquerre,fillDret)<0 ? fillEsquerre : fillDret;
	}


}
