package uoc.ei.tads;
/**
 * Seqncia que es caracteritza perqu es consulta i s'esborra el primer
 * element inserit: first-in-first-out (FIFO).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 * @inv n >= 0 && (n == darrer-primer+1 || n == max+darrer-primer+1)
 */
public class CuaVectorImpl<E>  implements Cua<E>, ContenidorAfitat<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Capacitat mxima, per defecte, del contenidor. */
   public static final int MAXIM_ELEMENTS_PER_DEFECTE = 256;

   /** Taula d'elements del contenidor. */
   protected E[] elements;

   /** Nombre d'elements que hi ha actualment al contenidor. */
   protected int n;

   /** Primer element de la cua. */
   private int  primer;

   /**
    * Constructor sense parmetres (capacitat mxima, per defecte).
    * @post n == 0 && this.max == MAXIM_ELEMENTS_PER_DEFECTE
    */
   public CuaVectorImpl()
   {
   	  this(MAXIM_ELEMENTS_PER_DEFECTE);
   }

   /**
    * Constructor amb un parmetre.
    * @exception ExcepcioParametreIncorrecte  si la capacitat mxima de la
    *                                         nova cua s negativa
    * @param max  nombre mxim d'elements de la cua 
    * @pre max>=0
    * @post n == 0 && this.max == max
    */
   public CuaVectorImpl(int max)
   {
      elements = (E[])new Object[max];
      n = 0;
      primer = 0;
   }
   
   
   private int posicio(int posicio) { return posicio % elements.length; }
   private int seguent(int posicio) { return (posicio+1)==elements.length ? 0 : posicio+1; }

   
   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    * @post $return == n && this == $old(this)
    */
   public int nombreElems() { return  n; }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    * @post $return == (n == 0) && this == $old(this)
    */
   public boolean estaBuit() { return  ( n == 0 ); }

   /**
    * Mtode per a comprovar si el contenidor est ple.
    * @return  cert o fals, segons si est ple o no ho est
    * @post $return == (n == max) && this == $old(this)
    */
   public boolean estaPle() { return  ( n == elements.length ); }

   /**
    * Afegeix un element a la cua, si hi cap.
    * @exception ExcepcioContenidorPle  si la cua est plena
    * @param elem  element que es vol afegir a la cua
    * @pre !estaPle(), ExcepcioContenidorPle
    * @post n == $old(n)+1 
    *  && primer == $old(primer)
    *  && darrer == ($old(darrer) == max-1 ? 0 : $old(darrer)+1)
    *  && vector[darrer] == elem
    *  && $all(i:vector,i == darrer || $old(vector[i]) == vector(i))
    */
   public void encuar(E elem)
   {
	  if (estaPle()) throw new ExcepcioContenidorPle();
   	  int darrer = posicio(primer + n);
      elements[darrer] = elem;
   	  n++;
   }

   /**
    * Esborra el primer element de la cua, si n'hi ha algun.
    * @exception ExcepcioContenidorBuit  si la cua est buida
    * @return  primer element inserit a la cua
    * @pre !estaBuit(), ExcepcioContenidorBuit
    * @post n == $old(n)-1
    *  && darrer == $old(darrer)
    *  && primer == ($old(primer) == max-1 ? 0 : $old(primer)+1)
    *  && $all(i:vector,i == $old(primer) || $old(vector[i]) == vector(i))
    */
   public E desencuar()
   {
	  if (estaBuit()) throw new ExcepcioContenidorBuit();
      E elem = elements[primer];     			 // primer element afegit
      elements[primer]=null;					 // necessari per al Garbage Collector
      primer = seguent(primer);					 // quan arriba a max torna a zero
      n--;                               		 // descompta l'element
      return  elem;                      		 // retorna l'element desencuat
   }

   /**
    * Accessor de lectura del primer element afegit a la cua, si n'hi ha.
    * @exception ExcepcioContenidorBuit  si la cua est buida
    * @return  primer element de la cua
    * @pre !estaBuit()
    * @post $return == vector[primer] && this == $old(this)
    */
   public E primer()
   {
      return  elements[primer];                    // retorna primer element
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor
    */
   public Iterador<E> elements() {
		return new IteradorVectorImpl<E>(elements,nombreElems(),primer);
}

   /**
    * Mtode que sobreescriu Object.toString(). Treu els elements separats
    * pel salt de lnia de la plataforma.
    * @return  llistat dels elements
    */
   public String toString()
   {
    StringBuffer buffer = new StringBuffer();
    buffer.append("{CUA:");
    for (Iterador<E> it = elements(); it.hiHaSeguent(); ) {
       buffer.append(it.seguent());
       if (it.hiHaSeguent()) buffer.append(',');
    }
    buffer.append("}");
    return  buffer.toString();
   }
}
