package uoc.ei.tads;
/**
 * Classe que implementa les operacions d'un diccionari per delegaci en
 * un arbre binari AVL.
 *
 * Els diccionaris sn estructures que emmagatzemen elements amb una clau
 * associada. La clau ha de disposar d'una operaci d'igualtat. En el cas
 * dels diccionaris ordenats ha de permetre l'ordenaci total entre claus.
 * Algunes implementacions permeten claus repetides i altres no. L'element
 * associat a la clau pot ser qualsevol objecte.
 *
 * En aquesta implementaci s'utilitzen elements de la classe ClauValor,
 * que emparella la clau i el valor associat. Si no hi ha comparador
 * especfic s'utilitza ClauValorComparable que implementa la interfcie
 * java.lang.Comparable, per delegaci en la clau.
 *
 * @see  ClauValor
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class DiccionariAVLImpl<C,E>  implements Diccionari<C,E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Arbre binari AVL que implementa les operacions per delegaci. */
   protected ArbreAVL<ClauValor<C,E>> avl;

   /**
    * Constructor sense parmetres. S'espera que les classes dels elements
    * implementin la interfcie java.lang.Comparable.
    */
   public DiccionariAVLImpl() { avl = new ArbreAVL<ClauValor<C,E>>(); }

   /**
    * Constructor amb un parmetre i elements d'una classe comparable amb
    * el comparador donat.
    * @param comparador  comparador que permet deduir la prioritat
    * @exception ExcepcioParametreIncorrecte  si el comparador s null
    */
   public DiccionariAVLImpl(java.util.Comparator<ClauValor<C,E>> comparador)
          throws  ExcepcioParametreIncorrecte
   {
      avl = new ArbreAVL<ClauValor<C,E>>(comparador);
   }

   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  avl.nombreElems(); }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  avl.estaBuit(); }

   /**
    * Afegeix un element amb una clau associada, si es pot. Si troba un
    * element amb la mateixa clau el sobreescriu.
    * @see  ClauValor
    * @param clau  clau associada a l'element que es vol afegir
    * @param elem  element que es vol afegir al diccionari
    */
   public void afegir(C clau,E elem)
   {
      avl.afegir( new ClauValor<C,E>(clau, elem));
   }

   /**
    * Comprova si hi ha un element amb una determinada clau.
    * @param clau  clau associada a un element
    * @return  cert o fals, segons si troba o no troba la clau
    */
   public boolean hiEs(C clau)
   {
      return  consultar(clau) != null;
   }

   /**
    * Accesor de lectura de l'element associat amb una clau.
    * @see  ClauValor
    * @param clau  clau de referncia
    * @return  element associat amb la clau; o null, si no hi era
    */
   public E consultar(C clau)
   {
   	E result=null;
   	ClauValor<C,E> aux=new ClauValor<C,E>(clau,null);
    ClauValor<C,E> clauValor =  avl.consultar(aux);
    if (clauValor!=null)
    	result=clauValor.getValor();
    return result;
   }

   /**
    * Esborra la primera clau coincident i l'element associat, si es pot.
    * @see  ClauValor
    * @param clau  clau de referncia
    * @return  element esborrat associat amb la clau;
    *          o null, si no hi era
    */
   public E esborrar(C clau)
   {
   	E result=null;
   	ClauValor<C,E> aux=new ClauValor<C,E>(clau,null);
    ClauValor<C,E> clauValor =  avl.esborrar(aux);
    if (clauValor!=null)
    	result=clauValor.getValor();
    return result;
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci de les claus del contenidor ordenades
    */
   public Iterador<C> claus()
   {
      return  new IteradorRecorregutClausImpl<C,E>(avl.recorregutInordre());
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements associats amb les claus
    */
   public Iterador<E> elements()
   {
      return  new IteradorRecorregutValorsImpl<C,E>(avl.recorregutInordre());
   }

   /**
    * Mtode que sobreescriu Object.toString(). Treu parelles d'elements.
    * Separa una parella de la segent amb el salt de lnia de la
    * plataforma.
    * @return  llistat de claus (entre claudtors) seguides de l'element
    *          associat a cada clau
    */
   public String toString() {
	   return  Utilitats.toStringContenidorDelegacio("DiccionariAVLImpl",avl);
   }
   
}
