package uoc.ei.tads;

import uoc.ei.tads.RecorregutBidireccional.IniciRecorregut;

/**
 * Seqncia posicional que es caracteritza per disposar d'operacions
 * basades en la posici que ocupa un element dins del contenidor.
 *
 * Hereta de llista encadenada l'estructura amb un sol encadenament (node
 * segent) i n'afegeix un segon per optimitzar les operacions que
 * necessiten conixer el node anterior dins de la seqncia.
 *
 * @see  Posicio
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class LlistaDoblementEncadenada<E>  extends LlistaEncadenada<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /**
    * Esborra la primera posici de la llista.
    * @exception ExcepcioContenidorBuit  si la llista est buida
    * @return  element que hi havia a la posici
    * @pre !estaBuit(), ExcepcioContenidorBuit
    */
   public E esborrarPrimer()  throws  ExcepcioContenidorBuit
   {
	  if (estaBuit()) throw new ExcepcioContenidorBuit();
   	  NodeDoblementEncadenat<E> primer =(NodeDoblementEncadenat<E>)darrer.getSeguent();
   	  if (n==1)
   	  	darrer=null;
   	  else {
        NodeDoblementEncadenat<E> segon =(NodeDoblementEncadenat<E>)primer.getSeguent();
        eliminarElDelMig((NodeDoblementEncadenat<E>)darrer,primer,segon);
      }
   	  n--;
   	  return primer.getElem();
   }

   
   /**
    * Esborra la posici rebuda.
    * @param node  posici que es vol eliminar
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  element que hi havia a la posici
    * @pre !estaBuit() && node!=null, ExcepcioPosicioInvalida
    */
   public E esborrar(Posicio<E> node)
   {
      NodeDoblementEncadenat<E> actual = (NodeDoblementEncadenat<E>)node;
   	  if (n==1)
   	  	darrer=null;
   	  else {
         NodeDoblementEncadenat<E> seguent=(NodeDoblementEncadenat<E>)actual.getSeguent();
         NodeDoblementEncadenat<E> anterior = actual.getAnterior();
         eliminarElDelMig(anterior,actual,seguent);
      }
   	  n--;
      return actual.getElem();
   }
   

   /** Esborra la posici segent.
    * @param node posici anterior a la que es vol eliminar; si s null
    * s'elimina la primera posici.
    * @pre !estaBuit() && node!=darrer, ExcepcioPosicioInvalida
    * @exception ExcepcioPosicioInvalida  si la posici s no vlida
    * @return  element que hi havia a la posici segent
    */
   public E esborrarSeguent(Posicio<E> node) {
	   if (node==darrer)
		   throw new ExcepcioPosicioInvalida();
   	  E elementEsborrat;
      if (node == null)
      	elementEsborrat=esborrarPrimer();
      else {
         NodeDoblementEncadenat<E> act = (NodeDoblementEncadenat<E>)node;
         NodeDoblementEncadenat<E> seg = (NodeDoblementEncadenat<E>)act.getSeguent();
         NodeDoblementEncadenat<E> segSeg = (NodeDoblementEncadenat<E>)seg.getSeguent();
         eliminarElDelMig(act,seg,segSeg);
         if (seg == darrer)
         	darrer = act;
         elementEsborrat=seg.getElem();
      }
      n--;
      return elementEsborrat;
   }

   
   /**
    * Retorna el node anterior al rebut com a parmetre. En aquesta llista
    * doblement encadenada l'operaci t un cost temporal u: O(1).
    * @param node  posici de referncia, no null
    * @return  node anterior al rebut com a parmetre
    */
   protected NodeEncadenat<E> anterior(NodeEncadenat<E> node)
   {
      return  ((NodeDoblementEncadenat<E>)node).getAnterior();
   }


   /** Mtode privat d'ajuda que donats tres nodes situats de manera consecutiva a la llista,
    * n'el.limina el del mig. Aquest mtode es fa servir per els mtodes d'esborrat de nodes.
    * El mtode nicament treu un element de la llista per no actualitza el nombre d'elements
    * d'aquesta.
    * @param anterior Element anterior.
    * @param central Element central, s l'element a el.liminar de la llista.
    * @param posterior Element posterior.
    */ 
   
   private void eliminarElDelMig(NodeDoblementEncadenat<E> anterior,NodeDoblementEncadenat<E> central,NodeDoblementEncadenat<E> posterior) {
   		// el.limina l'element central de la llista
	    anterior.setSeguent(posterior);
	    posterior.setAnterior(anterior);
	    // el.limina qualsevol referncia de l'element central a un element de la llista
	    central.setAnterior(null);
	    central.setSeguent(null);
	    if (central==darrer)
	    	darrer=anterior;
   }
   

   /**
    * Crea un node encadenat. Es pot sobreescriure per les subclasses que
    * necessitin treballar amb descendents de NodeEncadenat.
    * @param node  posici anterior a la del nou node; pot ser nul
    * @param elem  element que es vol afegir a la llista
    * @return  node nou contenint l'element rebut com a parmetre; situat
    *          entre el node rebut com a parmetre i el seu segent
    */
   protected NodeEncadenat<E> novaPosicio(NodeEncadenat<E> node, E elem)
   {
      NodeDoblementEncadenat<E> nouNode = null;

      if (node == null)                 // ser primer, darrer i nic node
      {
         nouNode = new NodeDoblementEncadenat<E>(elem);
         nouNode.setSeguent(nouNode);
         nouNode.setAnterior(nouNode);
         darrer = nouNode;
      }
      else
      {
         NodeDoblementEncadenat<E> actual = (NodeDoblementEncadenat<E>)node;
         NodeDoblementEncadenat<E> seguent = (NodeDoblementEncadenat<E>)actual.getSeguent();
         nouNode = new NodeDoblementEncadenat<E>(seguent,elem,(NodeDoblementEncadenat<E>)node);
         node.setSeguent(nouNode);
         seguent.setAnterior(nouNode);
      }
      n++;                              // incrementa el nombre d'elements
      return  nouNode;
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * Si durant el recorregut s'altera l'estructura de posicions del
    * contenidor, els resultats sn imprevisibles.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @see  IteradorBidireccional#hiHaAnterior()
    * @see  IteradorBidireccional#anterior()
    * @param inici  on s'inicia el recorregut; valors vlids sn:
    *                - IniciRecorregut.PRINCIPI i
    *                - IniciRecorregut.FINAL
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor en un sentit
    */
   public IteradorBidireccional<E> elements(IniciRecorregut inici)
   {
      return  new IteradorRecorregutBidireccionalImpl<E,E>(posicions(inici));
   }

   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @see  RecorregutBidireccional
    * @return  enumeraci de les posicions del contenidor
    */
   public RecorregutBidireccional<E> posicions(IniciRecorregut inici)
   {
      return  new RecorregutBidireccionalLlista<E>(this,inici);
   }


   /**
    * Classe que implementa un node amb dos encadenaments a node. La
    * interpretaci d'aquest parell d'encadenaments pot ser diversa: segent
    * / anterior en una llista doblement encadenada; fill esquerre / fill
    * dret en un arbre binari; etc. Noms facilita les operacions
    * bsiques: constructors i accessors de lectura i d'escriptura. Pot
    * emmagatzemar qualsevol element (Object).
    *
    * @author  Jordi lvarez Canal
    * @author  Esteve Marin Gallis.
    *          Estructura de la Informaci,
    *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
    */
   protected static class NodeDoblementEncadenat<E>  extends NodeEncadenat<E>
   {

      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
   	
   	
      /** Encadenament al node anterior. */
      protected NodeDoblementEncadenat<E> ant;


      /** Constructor sense parmetres. Dna valor nul als atributs. */
      public NodeDoblementEncadenat()
      {
      	  super();
      	  setAnterior(null);
      }

      
      /**
       * Constructor amb un parmetre.
       * @param elem  valor de l'element contingut al node
       */
      public NodeDoblementEncadenat(E elem)
      {
      	  super(elem);
      	  setAnterior(null);
      }

      
      /**
       * Constructor amb tres parmetres.
       * @param seg  encadenament al segent node
       * @param elem  valor de l'element contingut al node
       * @param ant encadenament al node anterior
       */
      public NodeDoblementEncadenat(NodeDoblementEncadenat<E> seg, E elem,
                                    NodeDoblementEncadenat<E> ant)
      {
      	  super(elem,seg);
      	  setAnterior(ant);
      }

      
      /**
       * Accessor d'escriptura del segon encadenament (node anterior en una
       * seqncia, fill dret en un arbre binari, etc.)
       * @param node  valor per al node 2
       */
      public void setAnterior(NodeDoblementEncadenat<E> node) { ant = node; }

      /**
       * Accesor de lectura del segon encadenament (node anterior en una
       * seqncia, fill dret en un arbre binari, etc.)
       * @return  segon encadenament a node
       */
      public NodeDoblementEncadenat<E> getAnterior() { return  ant; }
   }

   
   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions i s'actualitza d'acord amb els canvis.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    * @see  RecorregutBidireccional#hiHaAnterior()
    * @see  RecorregutBidireccional#anterior()
    */
   protected static class RecorregutBidireccionalLlista<E> extends RecorregutUnidireccionalLLista<E>
                                    implements RecorregutBidireccional<E>
   {
      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
      /** Hi pot haver anterior (control de l'encadenament circular). */
      protected boolean teAnterior;

      /**
       * Constructor amb un parmetre.
       * @param sentit  sentit del recorregut; valors vlids sn:
       *                -IteradorBidireccional.NORMAL i
       *                -IteradorBidireccional.INVERS
       */
      public RecorregutBidireccionalLlista(LlistaDoblementEncadenada<E> llista,IniciRecorregut sentit) {
         super(llista);
         if (sentit == IniciRecorregut.PRINCIPI)
          	teAnterior = false;
         else {
             teSeguent = false;
             if (darrer!=null)
             	nodeActual = darrer.getSeguent();
           	teAnterior=(nodeActual!=null);
          }
      }


      /**
       * Comprova si hi ha una ltima o anterior posici. s sensible a
       * eventuals alteracions de l'estructura de posicions. Retorna fals
       * si est buit o ja s'ha visitat la darrera posici. 
       * @return  cert o fals, segons si es pot recular o no es pot
       */
      public boolean hiHaAnterior()
      {
      	return  (darrer != null) && teAnterior;
      }

      /**
       * Primer avana, si es pot, i desprs retorna la posici.
       * @exception ExcepcioPosicioInvalida  si la segent posici no
       *            existeix
       * @return  segent posici
       * @pre hiHaSeguent(), ExcepcioPosicioInvalida
       */
      public Posicio<E> seguent() {
         teAnterior = true;
         return  super.seguent();
      }


      /**
       * Primer recula, si es pot, i desprs retorna la posici.
       * @exception ExcepcioPosicioInvalida  si l'anterior posici no
       *            existeix
       * @return  anterior posici
       * @pre hiHaAnterior(), ExcepcioPosicioInvalida
       */
      public Posicio<E> anterior() {
    	  if (!hiHaAnterior())
    		  throw new ExcepcioPosicioInvalida();
    	  nodeActual = ((NodeDoblementEncadenat<E>)nodeActual).getAnterior();
    	  // mirem si la posici anterior s la primera o no
    	  // amb l'objectiu de "marcar" si el nou node t anterior o no.
    	  teAnterior = (nodeActual != darrer.getSeguent());
    	  teSeguent = true;
    	  return  nodeActual;
      }
   }
}
