package uoc.ei.tads;
/**
 * Seqncia posicional que es caracteritza per disposar d'operacions
 * basades en la posici que ocupa un element dins del contenidor.
 *
 * @see  Posicio
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class LlistaEncadenada<E>  implements Llista<E>
{
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();


   /** Nombre d'elements que hi ha actualment al contenidor. */
   protected int n;

   /**
    * Posici del darrer element de la llista. En aquesta representaci
    * circular, la posici segent s la primera.
    */
   protected NodeEncadenat<E> darrer;


   /** Construeix una llista buida.
    */

   public LlistaEncadenada() {
	   darrer=null;
	   n=0;
   }


   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  n; }


   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  ( n == 0 ); }


   /**
    * Afegeix un element al principi de la llista.
    * @param elem  element que es vol afegir a la llista
    * @return  nova posici que cont l'element
    */
   public Posicio<E> afegirAlPrincipi(E elem) {
      return  novaPosicio(darrer, elem);
   }


   /**
    * Afegeix un element al final de la llista.
    * @param elem  element que es vol afegir a la llista
    * @return  nova posici que cont l'element
    */
   public Posicio<E> afegirAlFinal(E elem) {
      darrer = novaPosicio(darrer, elem);
	  return darrer;
   }


   /**
    * Afegeix un element abans de la posici rebuda.
    * @param elem  element que es vol afegir a la llista
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  nova posici que cont l'element
    */
   public Posicio<E> afegirAbansDe(Posicio<E> node, E elem) {
   	  Posicio<E> nouNode;
      NodeEncadenat<E> ant = anterior((NodeEncadenat<E>)node);
      if (ant==darrer)
      	nouNode=afegirAlPrincipi(elem);
      else
      	nouNode=afegirDespresDe(ant,elem);
      return nouNode;
   }


   /**
    * Afegeix un element desprs de la posici rebuda.
    * @param elem  element que es vol afegir a la llista
    * @param node  posici de referncia
    * @pre node!=null, ExcepcioPosicioInvalida
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  nova posici que cont l'element
    */
   public Posicio<E> afegirDespresDe(Posicio<E> node, E elem) {
   	  Posicio<E> nouNode;
   	  if (darrer==node)
   	  	nouNode=afegirAlFinal(elem);
   	  else
   	  	nouNode=novaPosicio((NodeEncadenat<E>)node,elem);
   	  return nouNode;
   }


   /**
    * Esborra la primera posici de la llista.
    * @pre estaBuit(), ExcepcioContenidorBuit
    * @exception ExcepcioContenidorBuit  si la llista est buida
    * @return  element que hi havia a la posici
    */
   public E esborrarPrimer() {
	   if (estaBuit())
		   throw new ExcepcioContenidorBuit();
      NodeEncadenat<E> primer = darrer.getSeguent();
      if (n == 1)
      	darrer = null;
      else
      	darrer.setSeguent( primer.getSeguent() );
      n--;
      return  primer.getElem();
   }


   /**
    * Esborra la posici rebuda.
    * @param node  posici que es vol eliminar
    * @pre node!=null, ExcepcioPosicioInvalida
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  element que hi havia a la posici
    */
   public E esborrar(Posicio<E> node) {
      if (n == 1)
      	darrer = null;
      else {
         NodeEncadenat<E> act = (NodeEncadenat<E>)node;
         NodeEncadenat<E> ant = anterior(act);
         ant.setSeguent(act.getSeguent());
         if (act == darrer)  darrer = ant;
      }
      n--;
      return  node.getElem();
   }


   /**
    * Esborra la posici segent.
    * @param node  posici anterior a la que es vol eliminar; si s null
    *              s'elimina la primera posici
    * @pre !estaBuit() && node!=darrer, ExcepcioPosicioInvalida
    * @exception ExcepcioPosicioInvalida  si la posici s no vlida
    * @return  element que hi havia a la posici segent
    */
   public E esborrarSeguent(Posicio<E> node) {
	   if (node==darrer)
		   throw new ExcepcioPosicioInvalida();
   	  E elementEsborrat;
      if (node == null)
      	elementEsborrat=esborrarPrimer();
      else {
         NodeEncadenat<E> act = (NodeEncadenat<E>)node;
         NodeEncadenat<E> seg = act.getSeguent();
         act.setSeguent(seg.getSeguent());
         if (seg == darrer)  darrer = act;  // si el segent era el darrer
         elementEsborrat=seg.getElem();
         n--;                                          // descompta l'element
      }
      return elementEsborrat;
   }


   /**
    * Reemplaa l'element contingut a la posici rebuda.
    * @param elem  nou element
    * @param node  posici de referncia
    * @exception ExcepcioPosicioInvalida  si la posici s null o no
    *            vlida
    * @return  element que hi havia a la posici
    * @pre !estaBuit() && node!=null, ExcepcioPosicioInvalida
    */
   public E reemplacar(Posicio<E> node, E elem) {
      E old = node.getElem();
      ((NodeEncadenat<E>)node).setElem(elem);
      return  old;
   }


   /**
    * Intercanvia els elements continguts a les posicions rebudes.
    * @param node1  primera de les dues posicions de referncia
    * @param node2  segona de les dues posicions de referncia
    * @exception ExcepcioPosicioInvalida  si alguna posici s null o no
    *            vlida
    * @pre !estaBuit() && node1!=null && node2!=null, ExcepcioPosicioInvalida
    */
   public void intercanviar(Posicio<E> node1, Posicio<E> node2) {
      E tmp = node1.getElem();
      ((NodeEncadenat<E>)node1).setElem(node2.getElem());
      ((NodeEncadenat<E>)node2).setElem(tmp);
   }


   /**
    * Retorna el node anterior al rebut com a parmetre. En aquesta llista
    * simplement encadenada l'operaci t un cost temporal lineal: O(n).
    * @param node  posici de referncia, no null
    * @return  node anterior al rebut com a parmetre
    * @pre nombreElems()>1
    */
   protected NodeEncadenat<E> anterior(NodeEncadenat<E> node) {
      NodeEncadenat<E> ant=darrer;
      boolean trobat=false;
      while (!trobat) {
      	trobat = ant.getSeguent()==node;
      	if (!trobat)
      		ant=ant.getSeguent();
      }
      return  ant;
   }


   /**
    * Crea un node encadenat, el situa a continuaci del rebut com a
    * parmetre i incrementa el nombre d'elements. Es pot sobreescriure
    * per les subclasses que necessitin treballar amb descendents de
    * NodeEncadenat.
    * @param node  posici anterior a la del nou node; pot ser null
    * @param elem  element que es vol afegir a la llista
    * @return  node nou contenint l'element rebut com a parmetre; situat
    *          entre el node rebut com a parmetre i el seu segent
    */
   protected NodeEncadenat<E> novaPosicio(NodeEncadenat<E> node, E elem) {
      NodeEncadenat<E> nouNode = null;
      if (node == null) {                // ser primer, darrer i nic node
         nouNode = new NodeEncadenat<E>(elem, null);
         nouNode.setSeguent(nouNode);
         darrer=nouNode;
      }
      else {
         nouNode = new NodeEncadenat<E>(elem, node.getSeguent());
         node.setSeguent(nouNode);
      }
      n++;                              // incrementa el nombre d'elements
      return  nouNode;
   }


   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * Si durant el recorregut s'altera l'estructura de posicions del
    * contenidor, els resultats sn imprevisibles.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor
    */
   public Iterador<E> elements() { return  new IteradorRecorregutImpl<E,E>(posicions()); }


   /**
    * Mtode que soporta mltiples recorreguts, de les posicions del
    * contenidor, simultanis i independents entre ells.
    * @return  enumeraci de les posicions del contenidor
    */
   public Recorregut<E> posicions() { return  new RecorregutUnidireccionalLLista<E>(this); }


   /**
    * Classe que implementa un node amb un encadenament a node, que uneix un
    * node amb el segent en una estructura de dades amb encadenament simple.
    * Noms facilita les operacions bsiques: constructors i accessors de
    * lectura i d'escriptura. Pot emmagatzemar qualsevol element (Object).
    *
    * @author  Jordi lvarez Canal
    * @author  Esteve Marin Gallis.
    *          Estructura de la Informaci,
    *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
    */
   protected static class NodeEncadenat<E>  implements Posicio<E> {

      /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
       * serialitzables de la mateixa classe. L'deitificador es calcula
       * mitjanant un mtode de la classe Utilitats. */
      private static final long serialVersionUID = Utilitats.getSerialVersionUID();


      /** Element contingut al node. */
      private E element;

      /** Encadenament al node segent. */
      protected NodeEncadenat<E> seg;

      /** Constructor sense parmetres. Dna valor nul als atributs. */
      public NodeEncadenat() {
      	  element = null;
      	  seg = null;
      }

      /**
       * Constructor amb un parmetre.
       * @param elem  valor de l'element contingut al node
       */
      public NodeEncadenat(E elem) {
         element = elem;
         seg = null;
      }

      /**
       * Constructor amb dos parmetres.
       * @param ne  encadenament a node
       * @param elem  valor de l'element contingut al node
       */
      public NodeEncadenat(E elem, NodeEncadenat<E> ne ) {
         element = elem;
         seg = ne;
      }

      /**
       * Accessor d'escriptura del valor contingut al node.
       * @param elem  nou valor de l'element contingut al node
       */
      public void setElem(E elem) { element = elem; }

      /**
       * Accessor de lectura de l'element contingut al node.
       * @return  element contingut al node
       */
      public E getElem() { return  element; }

      /**
       * Accessor d'escriptura de l'encadenament.
       * @param node  valor per al node
       */
      public void setSeguent(NodeEncadenat<E> node) { seg = node; }

      /**
       * Accesor de lectura del node encadenat.
       * @return  encadenament a node
       */
      public NodeEncadenat<E> getSeguent() { return  seg; }

      /**
       * Mtode que redefineix la conversi de l'objecte a String per
       * facilitar la depuraci del codi. Delega en el mtode toString() de
       * l'element emmagatzemat a la posici.
       * @return  cadena de carcters representativa de l'element
       */
      public String toString() {
         return  "[NodeEncadenat: "+((element == null) ? "null" : element.toString())+"]";
      }
   }


   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions i s'actualitza d'acord amb els canvis.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */

   protected static class RecorregutUnidireccionalLLista<E>  implements Recorregut<E> {

    /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
     * serialitzables de la mateixa classe. L'deitificador es calcula
     * mitjanant un mtode de la classe Utilitats. */
    private static final long serialVersionUID = Utilitats.getSerialVersionUID();


 	  /** Node actual. */
    protected NodeEncadenat<E> nodeActual;

 	  /** Darrer node de la llista. */
    protected NodeEncadenat<E> darrer;

      /** Hi pot haver segent (control de l'encadenament circular). */
      protected boolean teSeguent;


      public RecorregutUnidireccionalLLista(LlistaEncadenada<E> llista) {
      	nodeActual=llista.darrer;
      	darrer=llista.darrer;
      	teSeguent=(nodeActual!=null);
      }

      /**
       * Comprova si hi ha una primera o segent posici. s sensible a
       * eventuals alteracions de l'estructura de posicions. Retorna fals
       * si est buit o ja s'ha visitat l'ltima posici.
       * @return  cert o fals, segons si es pot avanar o no es pot
       */
      public boolean hiHaSeguent() { return  teSeguent; }

      /**
       * Primer avana, si es pot, i desprs retorna la posici.
       * @exception ExcepcioPosicioInvalida  si la segent posici no
       *            existeix
       * @return  segent posici
       * @pre hiHaSeguent(), new ExcepcioPosicioInvalida("no hi ha seguent")
       */
      public Posicio<E> seguent() {
    	  if (!hiHaSeguent())
    		  throw new ExcepcioPosicioInvalida();
         nodeActual = nodeActual.getSeguent();                 // actual <-- actual.segent
         teSeguent = (nodeActual != darrer);
         return  nodeActual;
      }
   }

   /**
    * Mtode que sobreescriu Object.toString(). Treu els elements separats
    * pel salt de lnia de la plataforma.
    * @return  llistat dels elements
    */
   public String toString() {
	  return Utilitats.toStringContenidor("LlistaEncadenada",new RecorregutUnidireccionalLLista<E>(this));
   }

}
