package uoc.ei.tads;
/**
 * Seqncia que es caracteritza perqu es consulta i s'esborra el darrer
 * element inserit: last-in-first-out (LIFO).
 *
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class PilaVectorImpl<E>  implements Pila<E>, ContenidorAfitat<E>
{
	
   /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
    * serialitzables de la mateixa classe. L'deitificador es calcula
    * mitjanant un mtode de la classe Utilitats. */
   private static final long serialVersionUID = Utilitats.getSerialVersionUID();
	
	
   /** Capacitat mxima, per defecte, del contenidor. */
   public static final int MAXIM_ELEMENTS_PER_DEFECTE = 256;

   /**
    * Nombre d'elements que hi ha actualment al contenidor. Tamb
    * representa la posici on s'ha d'empilar un nou element.
    */
   protected int n;

   /** Taula d'elements del contenidor. Les posicions comencen pel zero.*/
   protected E[] elements;

   /**
    * Constructor sense parmetres (capacitat mxima, per defecte).
    * @post n == 0 && this.max == MAXIM_ELEMENTS_PER_DEFECTE && elements.length==max
    */
   public PilaVectorImpl() {
   	  this(MAXIM_ELEMENTS_PER_DEFECTE);
   }

   /**
    * Constructor amb un parmetre.
    * @exception ExcepcioParametreIncorrecte  si la capacitat mxima de la
    *                                         nova pila s negativa
    * @param max  nombre mxim d'elements que pot contenir 
    * @pre max>0
    * @post n == 0 && this.max == max && elements.length==max
    */
   public PilaVectorImpl(int max) {
      elements = (E[])new Object[max];
      n = 0;
   }

   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    * @post $return==n && this == $old(this)
    */
   public int nombreElems() { return  n; }

   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    * @post $return==(n==0) && this == $old(this)
    */
   public boolean estaBuit() { return  ( n == 0 ); }

   /**
    * Mtode per a comprovar si el contenidor est ple.
    * @return  cert o fals, segons si est ple o no ho est
    * @post $return==(n==max) && this == $old(this)
    */
   public boolean estaPle() { return  ( n == elements.length ); }

   /**
    * Afegeix un element a la pila, si hi cap.
    * @exception ExcepcioContenidorPle  si la pila est plena
    * @param elem  element que es vol afegir a la pila
    * @pre !estaPle(), ExcepcioContenidorPle
    * @post n == $old(n)+1 
    * 		&& vector[n-1] == elem
    * 		&& $all(i:vector,i == n-1 || $old(vector[i]) == vector(i))
    */
   public void empilar(E elem) {
	   if (estaPle()) throw new ExcepcioContenidorPle();
      elements[n] = elem;
      n++;
   }

   /**
    * Esborra l'element del cim de la pila, si n'hi ha algun.
    * @exception ExcepcioContenidorBuit  si la pila est buida
    * @return  element que hi havia al cim de la pila
    * @pre !estaBuit(), ExcepcioContenidorBuit
    * @post n == $old(n)-1
    * 		&& $all(i:vector,i >= n || $old(vector[i]) == vector(i))
    */
   public E desempilar() {
	   if (estaBuit()) throw new ExcepcioContenidorBuit();
   		E aux=elements[n-1];
   		elements[n-1]=null;
   		n--;
   		return aux;
   }

   /**
    * Accessor de lectura del darrer element afegit a la pila, si n'hi ha.
    * @exception ExcepcioContenidorBuit  si la pila est buida
    * @return  element del cim de la pila
    * @pre !estaBuit()
    * @post $return == vector[n-1] && this==$old(this)
    */
   public E cim() {
      return  elements[n-1];
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @exception ExcepcioPosicioInvalida  si es vol obtenir el segent
    *            element de l'enumeraci i no n'hi ha cap o no n'hi ha cap
    *            ms
    * @return  enumeraci dels elements del contenidor
    */
   public Iterador<E> elements() {
   		return new IteradorVectorImpl<E>(elements,nombreElems(),0);
   }

   /**
    * Mtode que sobreescriu Object.toString(). Treu els elements separats
    * per comes.
    * @return Un String amb el llistat dels elements
    */
   public String toString()
   {
      StringBuffer buffer = new StringBuffer();
      buffer.append("{PILA:");
      for (Iterador it = elements(); it.hiHaSeguent(); ) {
         buffer.append(it.seguent());
         if (it.hiHaSeguent()) buffer.append(',');
      }
      buffer.append("}");
      return  buffer.toString();
   }
}
