package uoc.ei.tads;

/**
 * Classe que implementa les operacions d'un diccionari mitjanant una
 * taula de dispersi indirecta, coneguda amb el nom de taula encadenada
 * oberta (separate chaining). Segurament s la manera ms senzilla de
 * resoldre les collisions. Per la funci de dispersi s'accedeix a
 * l'ndex del vector i els sinnims s'encadenen a partir d'aquesta
 * posici.
 *
 * Els diccionaris sn estructures que emmagatzemen elements amb una clau
 * associada. La clau ha de disposar d'una operaci d'igualtat i l'element
 * associat a la clau pot ser qualsevol objecte.
 *
 * En aquesta implementaci s'utilitzen elements de la classe ClauValor
 * que sobreescriu, per delegaci en la clau, la funci hashCode().
 *
 * @see  ClauValor
 * @author  Jordi lvarez Canal
 * @author  Esteve Marin Gallis.
 *          Estructura de la Informaci,
 *          Universitat Oberta de Catalunya (UOC)
 * @version  2.0.0
 * @since  1.5
 */
public class TaulaDispersio<C,E>  implements Diccionari<C,E>
{
    /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
     * serialitzables de la mateixa classe. L'deitificador es calcula
     * mitjanant un mtode de la classe Utilitats. */
    private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
   /** Mida per defecte de la taula de dispersi. */
   public static final int MIDA_TAULA_PER_DEFECTE = 257;


   /** Nombre d'elements que hi ha actualment al contenidor. */
   protected int n = 0;

   
   /** Vector de nodes encadenats. */
   protected LlistaEncadenada<ClauValor<C,E>>[] taula;

   
   /** Constructor sense parmetres (mida de la taula per defecte). */
   public TaulaDispersio()
   {
   	  this(MIDA_TAULA_PER_DEFECTE);
   }

   
   /**
    * Constructor amb un parmetre. Crea un vector amb la mida donada.
    * @exception ExcepcioParametreIncorrecte  si la mida s negativa
    * @param mida  mida de la taula de dispersi
    * @pre mida>=0, new ExcepcioParametreIncorrecte("la mida no pot ser negativa")
    */
   public TaulaDispersio(int mida)  throws  ExcepcioParametreIncorrecte
   {
      taula = new LlistaEncadenada[mida];
   }

   
   /**
    * Accessor de lectura del nombre d'elements que hi ha al contenidor.
    * @return  nombre d'elements que cont actualment
    */
   public int nombreElems() { return  n; }

   
   /**
    * Mtode per a comprovar si el contenidor est buit.
    * @return  cert o fals, segons si est buit o no ho est
    */
   public boolean estaBuit() { return  (n == 0); }

   
   /**
    * Afegeix un element amb una clau associada, si es pot. Si troba un
    * element amb la mateixa clau el sobreescriu.
    * @see  ClauValor
    * @param clau  clau associada a l'element que es vol afegir
    * @param elem  element que es vol afegir al diccionari
    * @exception ExcepcioTADs  si la clau s null
    */
   public void afegir(C clau, E elem) {
      ClauValor<C,E> kv = new ClauValor<C,E>(clau, elem);
      int index = calcularIndexTaula(clau);
      LlistaEncadenada<ClauValor<C,E>> sinonims=taula[index];
      Posicio<ClauValor<C,E>> posicio=null;
      if (sinonims==null)
      	sinonims=creaLlistaDeSinonims(index,kv);
      else
      	posicio=cercarClauEnSinonims(sinonims,clau);
      if (posicio!=null)
    	  sinonims.reemplacar(posicio,kv);
      else {
    	  sinonims.afegirAlPrincipi(kv);
    	  n++;
      }
   }


   /**
    * Comprova si hi ha un element amb una determinada clau.
    * @param clau  clau associada a un element
    * @return  cert o fals, segons si troba o no troba la clau
    */
   public boolean hiEs(C clau) {
      return  consultar(clau) != null;
   }

   
   /**
    * Accesor de lectura de l'element associat amb una clau.
    * @see  ClauValor
    * @param clau  clau de referncia
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @return  element ClauValor associat amb la clau;
    *          o null, si no hi era
    */
   public E consultar(C clau) {
   		E objecteCercat=null;
		int index = calcularIndexTaula(clau);
		LlistaEncadenada<ClauValor<C,E>> sinonims=taula[index];
		Posicio<ClauValor<C,E>> posicioElement=cercarClauEnSinonims(sinonims,clau);
		if (posicioElement!=null) {
			ClauValor<C,E> kv=posicioElement.getElem();
			objecteCercat=kv.getValor();
		}
		return objecteCercat;
   }


   /**
    * Esborra la primera clau coincident i l'element associat, si es pot.
    * @see  ClauValor
    * @param clau  clau de referncia
    * @exception ExcepcioContenidorBuit  si la taula est buida
    * @exception ExcepcioElementsNoComparables  si l'element s no
    *            comparable
    * @return  element esborrat associat amb la clau;
    *          o null, si no hi era
    */
   public E esborrar(C clau)
   {
      E valorAntic = null;
      int index = calcularIndexTaula(clau);
      LlistaEncadenada<ClauValor<C,E>> sinonims=taula[index];
      if (sinonims!=null) {
	 	Recorregut<ClauValor<C,E>> recorregutSinonims=sinonims.posicions();
	  	Posicio<ClauValor<C,E>> posicio=null,posicioAnterior=null;
	  	boolean trobat=false;
	  	while (recorregutSinonims.hiHaSeguent() && !trobat) {
	  		posicioAnterior=posicio;
	  		posicio=recorregutSinonims.seguent();
	  		ClauValor<C,E> kvPosicio=posicio.getElem();
	  		trobat=clau.equals(kvPosicio.getClau());
	  	}
	  	if (trobat) {
		  	valorAntic=posicio.getElem().getValor();
	  		if (sinonims.nombreElems()==1)
	  			esborrarLlistaDeSinonims(index);
	  		else
			  	sinonims.esborrarSeguent(posicioAnterior);
		  	n--;
	  	}
      }
	 return valorAntic;
   }

   
   protected int calcularIndexTaula(C clau) {
	   int hash = clau.hashCode();                  // codi hash de la clau
	   // retalla el bit de signe i ajusta l'ndex a l'interval [0..max-1]: 
	   return (hash & 0x7FFFFFFF) % taula.length;
   }


   protected LlistaEncadenada<ClauValor<C,E>> creaLlistaDeSinonims(int index,ClauValor<C,E> kv) {
		taula[index] = new LlistaEncadenada<ClauValor<C,E>>();
		return taula[index];
   }


   protected void esborrarLlistaDeSinonims(int index) {
		taula[index] = null;
   }


   protected Posicio<ClauValor<C,E>> cercarClauEnSinonims(LlistaEncadenada<ClauValor<C,E>> sinonims,C clau) {
	  	Posicio<ClauValor<C,E>> posicio=null;
	  	boolean trobat=false;
		if (sinonims!=null) {
		 	Recorregut<ClauValor<C,E>> recorregutSinonims=sinonims.posicions();
		  	while (recorregutSinonims.hiHaSeguent() && !trobat) {
		  		posicio=recorregutSinonims.seguent();
		  		ClauValor<C,E> kvPosicio=posicio.getElem();
		  		trobat=clau.equals(kvPosicio.getClau());
		  	}
		}
	  	return trobat ? posicio : null;
	}

	
   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @return  enumeraci de les claus del contenidor sense ordenar
    */
   public Iterador<C> claus() {
      return  new IteradorRecorregutClausImpl<C,E>(new RecorregutNodes<C,E>(this));
   }

   /**
    * Accessor de lectura dels elements que hi ha al contenidor.
    * Retorna una enumeraci. Es pot obtenir un llistat amb un parell de
    * lnies de codi: <PRE>
    * for ( Iterador it = tad.elements(); it.hiHaSeguent(); )
    *    System.out.println(it.seguent()); </PRE>
    * Enumerar s simplement enunciar l'una darrere l'altra (les coses
    * d'una srie, les parts d'un tot). Per si el contenidor t definit
    * algun tipus d'ordenaci o de recorregut, l'enumeraci ha de ser
    * conseqent i oferir els elements per ordre (FIFO, LIFO, inordre,
    * etc.), sense alterar l'estat actual del contenidor.
    * @see  Iterador#hiHaSeguent()
    * @see  Iterador#seguent()
    * @return  enumeraci dels elements associats amb les claus
    */
   public Iterador<E> elements() {
      return  new IteradorRecorregutValorsImpl<C,E>(new RecorregutNodes<C,E>(this));
   }

   /**
    * Classe que proporciona un recorregut de les posicions. Basada en el
    * pattern Iterator, soporta mltiples recorreguts simultanis i
    * independents del contenidor. s sensible a eventuals alteracions de
    * l'estructura de posicions.
    * @see  Recorregut#hiHaSeguent()
    * @see  Recorregut#seguent()
    */
   protected static class RecorregutNodes<C,E>  implements Recorregut<ClauValor<C,E>> {
    /** Atribut que en Java 1.5 determina la compatibilitat entre objectes
     * serialitzables de la mateixa classe. L'deitificador es calcula
     * mitjanant un mtode de la classe Utilitats. */
    private static final long serialVersionUID = Utilitats.getSerialVersionUID();
 	
 	
    /** Taula de dispersi que s'est recorrent. */
    protected TaulaDispersio<C,E> taulaDispersio;

    /** Recorregut auxiliar. */
    protected Recorregut<ClauValor<C,E>> recorregutSinonims = null;

    /** ndex de la taula. */
    protected int i = 0;

    public RecorregutNodes(TaulaDispersio<C,E> td) {
    	taulaDispersio=td;
    	recorregutSinonims = null;
    	i = 0;
		cercarSeguent();
    }
    
	
    /**
     * Comprova si hi ha una primera o segent posici. s sensible a
     * eventuals alteracions de l'estructura de posicions. Retorna fals
     * si el contenidor est buit o ja s'ha visitat l'ltima posici.
     * @return  cert o fals, segons si es pot avanar o no es pot
     */
    public boolean hiHaSeguent() { return  recorregutSinonims!=null && recorregutSinonims.hiHaSeguent(); }

	
    /**
     * Primer avana, si es pot, i desprs retorna la posici.
     * Si no hi ha segent posici llana una excepci.
     * @pre hiHaSeguent()
     * @return  segent posici
     */
    public Posicio<ClauValor<C,E>> seguent() {
    	if (!hiHaSeguent())
    		throw new ExcepcioPosicioInvalida();
		Posicio<ClauValor<C,E>> posicio=recorregutSinonims.seguent();
		cercarSeguent();
		return posicio;
    }

	
    private void cercarSeguent() {
		while (i<taulaDispersio.taula.length && (recorregutSinonims==null || !recorregutSinonims.hiHaSeguent())) {
			if (taulaDispersio.taula[i]!=null)
				recorregutSinonims=taulaDispersio.taula[i].posicions();
			else
				recorregutSinonims=null;
			i++;
		}
    }
    
   }

   
   /**
    * Mtode que sobreescriu Object.toString(). Treu parelles d'elements.
    * Separa una parella de la segent amb el salt de lnia de la
    * plataforma.
    * @return  llistat de claus (entre claudtors) seguides de l'element
    *          associat a cada clau
    */
   public String toString()
   {
   	  return Utilitats.toStringContenidor("TaulaDispersio",new RecorregutNodes<C,E>(this));
   }
   
}
